"""
Icom IC-F621 Land Mobile Radio
"""

# Copyright 2023 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from chirp import chirp_common
from chirp import directory
from chirp import bitwise
from chirp.drivers import icf
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueInteger, RadioSettingValueList, \
    RadioSettingValueString, RadioSettingValueFloat

LOG = logging.getLogger(__name__)

MEM_FORMAT = """
u8    model_code[4];
u8    unknown100[269];
u8    beep_on:1,
      beep_level_linked:1,
      beep_level:6;
u8    squelch_level;
u8    volume_min_level;
u8    unknown101:5,
      mic_gain:3;
u8    unknown102[9];
u8    unknown103:3,
      ignition_switch_on:1,
      audio_filter:2,
      backlight:2;
u8    horn_off:1,
      unknown104:7;
u8    unknown105[16];
struct {
  u16   user_1:4,
        user_2:4,
        user_3:4,
        user_4:4;
  u24   dealer_1:4,
        dealer_2:4,
        dealer_3:4,
        dealer_4:4,
        dealer_5:4,
        dealer_6:4;
} passwd;
u8    scrambler_syn_capture:2,
      scrambler_tone_start_timing:2,
      scrambler_non_rolling:1,
      unknown106:1,
      scrambler_group_code:2;
u8    unknown107[4];
u16   beep_frequency_low;
u16   beep_frequency_high;
u16   unknown107a;
u8    backlight_set_enable:1,
      beep_set_enable:1,
      beep_level_set_enable:1,
      squelch_level_set_enable:1,
      volume_min_level_set_enable:1,
      audio_filter_set_enable:1,
      mic_gain_set_enable:1,
      horn_set_enable:1;
u8    unknown108[5];
bit   trunk_on[16];
struct {
  u8    emergency;
  u8    priority_a;
  u8    priority_b;
} channel;
struct {
  u8    unk1;
  u8    unk2;
  u8    unk3;
  u8    unk4;
  u8    unk5;
  u8    unk6;
} unknown109;
u8    tx_dtcs_inverse:1,
      rx_dtcs_inverse:1,
      unknown110:6;
u8    unknown111:2,
      emergency_channel_on:1,
      unknown112:1,
      tone_burst_phase_on:1,
      rx_af_filter_low_cut_on:1,
      unknown113:2;
u8    poweron_scan_on:1,
      unknown114:1,
      dtmf_id_out_on:1,
      time_out_timer_beep_on:1,
      unknown115:1,
      data_out_on:1,
      poweron_priority_a:1,
      unknown116:1;
u8    unknown117:2,
      opening_text_on:1,
      poweron_selection_off:1,
      unknown118:4;
u8    unknown119:2,
      password_on:1,
      unknown120:2,
      rf_power_selection:1,
      unknown121:2;
u8    unknown122;
u8    hookoff_monitor_on:1,
      hookoff_priority_a:1,
      hookon_scan_on:1,
      unknown123:1,
      exo_on:1,
      tone_mute_eptt_on:1,
      unknown124:2;
struct {
  u8    scan_fast;
  u8    scan_slow;
  u8    ctcss_reverse_burst;
  u8    unk1;
  u8    unk2;
  u8    unk3;
  u8    unk4;
  u8    two_tone_1_period;
  u8    unk5;
  u8    dtmf;
  u8    dtmf_first;
  u8    dtmf_starpound;
  u8    two_tone_group;
  u8    scan_stop;
  u8    scan_resume;
  u8    unk6;
  u8    unk7;
  u8    emergency_switch_on;
  u8    emergency_switch_off;
  u8    emergency_start_repeat;
  u8    time_out;
  u8    penalty;
  u8    lockout_penalty;
  u8    two_tone_beep_repeat;
  u8    auto_reset_a;
  u8    auto_reset_b;
  u8    unk8;
  u8    two_tone_auto_tx;
  u8    horn;
  u8    exo_delay;
  u8    eptt_delay;
  u8    ignition_switch_delay:4,
        unk9:4;
} timer;
struct {
  u8    lup;
  u8    ldown;
  u8    rup;
  u8    rdown;
  u8    p0;
  u8    p1;
  u8    p2;
  u8    p3;
  u8    p4;
  u8    opf0;
  u8    opf1;
  u8    opf2;
} conv_key;
struct {
  u8    lup;
  u8    ldown;
  u8    rup;
  u8    rdown;
  u8    p0;
  u8    p1;
  u8    p2;
  u8    p3;
  u8    p4;
  u8    opf0;
  u8    opf1;
  u8    opf2;
} trunk_key;
struct {
  u16   capacity;
} bank[16];
struct {
  u16   unk;
} unknown125[16];
char clone_comment[32];
struct {
  u16   map;
} custom_characters[16];
u16   two_tone_1_tone;
u16   two_tone_2_tone;
struct {
  u16   unk;
} unknown126[40];
u8    two_tone_2_period;
struct {
  u8    unk;
} unknown127[14];
u8    two_tone_timer_notone;
struct {
  u8    unk;
} unknown128[22];
u16   user_ctcss_freq;
u32   unknown129;
struct {
  u16   rx;
  u16   tx;
} continuous_tone[9];
struct {
  char  name[10];
  bit   active[256];
  u8    text_on:1,
        unk1:5,
        scan_mode:2;
  u8    unk2;
} scan_list[10];
struct {
  char  name[10];
  u8    code[12];
  u8    count;
} dtmf_autodial[10];
struct {
  char  name[10];
  u32   disp_inhibit:1,
        freq_rx:31;
  u32   tx_inhibit:1,
        freq_tx:31;
  u16   rx_tone_off:1,
        rx_tone_digital:1,
        unk01:5,
        rx_tone:9;
  u16   tx_tone_off:1,
        tx_tone_digital:1,
        unk02:5,
        tx_tone:9;
  u8    unk03:3,
        tot_on:1,
        lockout_repeater:1,
        lockout_busy:1,
        power_rf:2;
  u8    log_in:2,
        log_out:2,
        unk04:1,
        text_on:1,
        unk05:1,
        two_tone_unk1:1;
  u8    unk06:4,
        two_tone_unk2:2,
        auto_reset:1,
        unk07:1;
  u8    narrow:1,
        scrambler_on:1,
        scrambler_inhibit:1,
        compander_on:1,
        unk08:4;
  u8    unk09;
  u8    scrambler_code;
  u16   unk10;
  u16   unk11;
  u8    unk12:6,
        two_tone_index:2;
} memory[256];
struct {
  char  name[10];
  u16   unk1;
  u16   unk2;
  u16   unk3;
} unknown130[32];
u32   unknown131;
struct {
  char name[10];
  u8    unk1;
  u8    ans_on;
  u8    unk2:2,
        scan_start:1,
        scan_cancel:1,
        unk3:1,
        beep:3;
  u8    exo_on:1,
        unk4:1,
        auto_tx_on:1,
        stun_on:1,
        kill_on:1,
        unk5:1,
        bell:2;
  u8    unk6;
  u16   unk7;
  u16   unk8;
} two_tone[9];
u8    unknown132[4];
char  opening_text[10];

"""

APPEAR_CHOICE = ["Appear", "Inhibit"]
ENABLE_CHOICE = ["Disable", "Enable"]
INVERSE_CHOICE = ["Normal", "Inverse"]
NO_YES_CHOICE = ["No", "Yes"]
ON_CHOICE = ["", "On"]          # ["OFF", "ON"]
ON_OFF_CHOICE = ["Off", "On"]
YES_CHOICE = ["", "Yes"]

MODES = ["FM", "NFM"]           # Supported emission (or receive) modes

TMODES = ["", "Tone", "TSQL", "DTCS", "Cross", ]
CROSS_MODES = [
    "Tone->Tone",
    "Tone->DTCS",
    "DTCS->Tone",
    "DTCS->",
    "->Tone",
    "->DTCS",
    "DTCS->DTCS",
    ]

TONES = [
    67.0, 69.3, 71.0, 71.9, 74.4, 77.0, 79.7, 82.5,
    85.4, 88.5, 91.5, 94.8, 97.4, 100.0, 103.5,
    107.2, 110.9, 114.8, 118.8, 123.0, 127.3,
    131.8, 136.5, 141.3, 146.2, 151.4, 156.7,
    159.8, 162.2, 165.5, 167.9, 171.3, 173.8,
    177.3, 179.9, 183.5, 186.2, 189.9, 192.8,
    196.6, 199.5, 203.5, 206.5, 210.7, 218.1,
    225.7, 229.1, 233.6, 241.8, 250.3, 254.1,
    ]

DUPLEX = ["", "+", "-", "split", "off"]

# Valid values for programmable keys.
# Hex numbers are used for unknown functions.
KEY_FUNC = [
    "Null",
    "CH Up",
    "CH Down",
    "Bank",
    "Scan A Start/Stop",
    "Scan B Start/Stop",
    "Scan Add/Del(Tag)",
    "Prio A",
    "Prio A(Rewrite)",
    "Prio B",
    "MR-CH 1",
    "MR-CH 2",
    "MR-CH 3",
    "MR-CH 4",
    "Moni",
    "Lock",
    "High/Low",
    "Wide/Narrow",
    "Compander",
    "C.Tone CH Ent",
    "Talk Around",
    "Scrambler",
    "Re-dial",
    "DTMF Autodial",
    "Call",
    "0x19",
    "0x1A",
    "Emergency Single",
    "Emergency Repeat",
    "0x1D",
    "0x1E",
    "0x1F",
    "0x20",
    "User Set Mode",
    "Public Address",
    "RX Speaker",
    "0x24",
    "0x25",
    "0x26",
    "0x27",
    "Turbo SpeeDial A",
    "Turbo SpeeDial B",
    "Turbo SpeeDial C",
    "Turbo SpeeDial D",
    "Trunking Group Switch",
    "Sp. Func 1",
    "Sp. Func 2",
    "Hook Scan",
    "OPT1 Out",
    "OPT2 Out",
    "OPT3 Out",
    "OPT1 Momentary",
    "OPT2 Momentary",
    "OPT3 Momentary",
    ]

# List of valid conventional keys
# in the order they should be presented to the user.
KEY_CONV = [
    "Null",
    "CH Up",
    "CH Down",
    "Bank",
    "Scan A Start/Stop",
    "Scan B Start/Stop",
    "Scan Add/Del(Tag)",
    "Prio A",
    "Prio A(Rewrite)",
    "Prio B",
    "MR-CH 1",
    "MR-CH 2",
    "MR-CH 3",
    "MR-CH 4",
    "Moni",
    "Public Address",
    "RX Speaker",
    "Lock",
    "High/Low",
    "C.Tone CH Ent",
    "Talk Around",
    "Wide/Narrow",
    "DTMF Autodial",
    "Re-dial",
    "Call",
    "Emergency Single",
    "Emergency Repeat",
    "Scrambler",
    "Compander",
    "Hook Scan",
    "User Set Mode",
    "OPT1 Out",
    "OPT2 Out",
    "OPT3 Out",
    "OPT1 Momentary",
    "OPT2 Momentary",
    "OPT3 Momentary",
    "Sp. Func 1",
    "Sp. Func 2",
    ]

# List of valid trunk keys
# in the order they should be presented to the user.
KEY_TRUNK = [
    "Null",
    "Bank",
    "Lock",
    "High/Low",
    "Trunking Group Switch",
    "Turbo SpeeDial A",
    "Turbo SpeeDial B",
    "Turbo SpeeDial C",
    "Turbo SpeeDial D",
    "CH Up",
    "CH Down",
    "Compander",
    "Sp. Func 1",
    "Sp. Func 2",
    ]

AUDIO_FILTER_RX_OPTIONS = [
    "Normal",
    "Low Cut",
    ]

RF_POWER_SELECTION_OPTIONS = [
    "MR CH Individual",
    "Override",
    ]

BACKLIGHT_VALUES = [
    "Off",
    "Dim",
    "Auto",
    "On",
    ]

BEEP_LEVEL_VALUES = [
    "",
    "1",
    "2",
    "3",
    "4",
    "5",
    ]

AUDIO_FILTER_VALUES = [
    "300-3000 Hz",
    "0-3000 Hz",
    "300-3400 Hz",
    "0-3400 Hz",
    ]

# Valid values for Mic Gain.
# Hex numbers are used for unknown functions.
MIC_GAIN_FUNC = [
    "0x00",
    "1 Min",
    "2",
    "3",
    "4",
    "5 Max",
    ]

# List of valid Mic Gain choices
# in the order they should be presented to the user.
MIC_GAIN_CHOICE = [
    "1 Min",
    "2",
    "3",
    "4",
    "5 Max",
    ]

LOCKOUT_VALUES = [              # CS-F500 equivalent
    "",                         # OFF
    "Repeater",                 # Repeater Lockout
    "Busy",                     # Busy Lockout
    ]

LOG_IN_OUT_VALUES = [           # CS-F500 equivalent
    "",                         # OFF
    "L-IN",                     # Log IN
    "L-OFF",                    # Log OFF
    "Both",                     # Log IN/OFF
    ]

RESET_TIMERS_VALUES = [
    "Timer B",
    "Timer A",
    ]

SCRAMBLER_TYPE_VALUES = [
    "Rolling",
    "Non-rolling",
    ]

SCRAMBLER_GCODE_VALUES = [
    "1",
    "2",
    "3",
    "4",
    ]

# Valid values for Scrambler Synchronous Capture.
# Hex numbers are used for unknown functions.
SYN_CAPTURE_FUNC = [
    "Standard",
    "0x01",
    "0x02",
    "Continuous",
    ]

# List of valid Scrambler Synchronous Capture choices
# in the order they should be presented to the user.
SYN_CAPTURE_CHOICE = [
    "Standard",
    "Continuous",
    ]

TONE_BURST_VALUES = [
    "Notone",
    "Phase",
    ]

TONE_STIMING_VALUES = [
    "Off",
    "300 ms",
    "600 ms",
    "1100 ms",
    ]


def kf_kc(i):
    """
    Given the index of the conventional key choice
    return the index of the corresponding function
    """
    return int(KEY_FUNC.index(KEY_CONV[int(i)]))


def kc_kf(i):
    """
    Given the index of the key function
    return the index of the corresponding conventional key choice
    """
    return int(KEY_CONV.index(KEY_FUNC[int(i)]))


def kf_kt(i):
    """
    Given the index of the trunk key choice
    return the index of the corresponding function
    """
    return int(KEY_FUNC.index(KEY_TRUNK[int(i)]))


def kt_kf(i):
    """
    Given the index of the key function
    return the index of the corresponding trunk key choice
    """
    return int(KEY_TRUNK.index(KEY_FUNC[int(i)]))


def mf_mc(i):
    """
    Given the index of the mic gain choice
    return the index of the corresponding function
    """
    return int(MIC_GAIN_FUNC.index(MIC_GAIN_CHOICE[int(i)]))


def mc_mf(i):
    """
    Given the index of the mic gain function
    return the index of the corresponding choice
    """
    return int(MIC_GAIN_CHOICE.index(MIC_GAIN_FUNC[int(i)]))


def sf_sc(i):
    """
    Given the index of the sync capture choice
    return the index of the corresponding function
    """
    return int(SYN_CAPTURE_FUNC.index(SYN_CAPTURE_CHOICE[int(i)]))


def sc_sf(i):
    """
    Given the index of the sync capture function
    return the index of the corresponding choice
    """
    return int(SYN_CAPTURE_CHOICE.index(SYN_CAPTURE_FUNC[int(i)]))


@directory.register
class ICF621_2Radio(icf.IcomCloneModeRadio):
    """
    Icom IC-F621-2 UHF Land Mobile Radio
    """
    VENDOR = "Icom"
    MODEL = "IC-F621-2"

    # _power_high watts is taken from the brochure.
    # FIXME: The others are guesses
    _power_low1 = chirp_common.PowerLevel("Low1", watts=5)
    _power_low2 = chirp_common.PowerLevel("Low2", watts=20)
    _power_high = chirp_common.PowerLevel("High", watts=45)

    _model = "\x25\x26\x02\x00"     # 4 byte model (don't use Byte String)
    _memsize = 0x3000
    _endframe = "Icom Inc."         # \x49\x63\x6F\x6D\x20\x49\x6E\x63\x2E

    # Ranges of the mmap to send to the radio
    _ranges = [(0x0100, 0x3000, 16)]

    _num_banks = 16
    _can_hispeed = False

    # Different models and versions support subsets of the valid bands.
    # Frequency values out of range for a radio will not cause a
    # transfer error.  The radio display will flash when that channel
    # is selected.
    # IC-F521 _valid_bands = [(136000000, 174000000)]
    _valid_bands = [(440000000, 490000000)]

    _memories = {}

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True

        # Channels
        rf.has_bank = False
        rf.has_bank_names = False
        rf.memory_bounds = (1, 256)

        # Channel Name
        rf.has_name = True
        rf.valid_characters = chirp_common.CHARSET_ASCII
        rf.valid_name_length = 10

        # Frequency
        rf.valid_modes = list(MODES)
        rf.has_offset = True
        rf.has_tuning_step = False
        rf.valid_tuning_steps = chirp_common.TUNING_STEPS
        rf.valid_duplexes = list(DUPLEX)
        rf.valid_bands = list(self._valid_bands)
        rf.can_odd_split = True
        rf.has_mode = True

        # Tone
        rf.has_ctone = True
        rf.has_dtcs = True
        rf.has_rx_dtcs = True
        rf.has_dtcs_polarity = False
        rf.has_cross = True
        rf.valid_cross_modes = list(CROSS_MODES)
        rf.valid_tmodes = list(TMODES)
        rf.valid_tones = list(TONES)

        # RF Power
        rf.has_variable_power = False
        rf.valid_power_levels = [
            self._power_high, self._power_low2, self._power_low1]

        # Scanning
        rf.valid_skips = []

        return rf

    def process_mmap(self):
        """Convert the raw byte array into a memory object structure"""
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

        # FIXME: BEGIN Ugly Hack
        # Force driver to use one bank that contains all memories.
        # This allows us to work on other parts of the driver before
        # the bank model is complete.
        self._memobj.bank[0].capacity = 256
        for i in range(1, self._num_banks):
            self._memobj.bank[i].capacity = 0
        # When deleting this ugly hack, remember to change:
        #    rf.has_bank = True
        # below.
        # FIXME: END Ugly Hack

    def get_memory(self, number):
        """Create a high-level memory object to return to the UI"""

        _mem = self._memobj.memory[number - 1]
        memory = chirp_common.Memory()
        memory.number = number
        memory.immutable = []

        # Frequency: Low level format is frequency in Hz
        memory.freq = int(_mem.freq_rx)

        if _mem.tx_inhibit:
            memory.duplex = "off"
            memory.offset = 0
        elif int(_mem.freq_rx) == int(_mem.freq_tx):
            memory.duplex = ""
            memory.offset = 0
        else:
            chirp_common.split_to_offset(
                memory, int(_mem.freq_rx), int(_mem.freq_tx))

        # Modes
        if _mem.narrow:
            memory.mode = "NFM"
        else:
            memory.mode = "FM"

        # Name
        memory.name = str(_mem.name).rstrip()

        # Tone
        txpol = None
        if _mem.tx_tone_off:
            txmode = None
            txval = None
        elif _mem.tx_tone_digital:
            txmode = "DTCS"
            txval = chirp_common.ALL_DTCS_CODES[int(_mem.tx_tone)]
        else:
            txmode = "Tone"
            txval = TONES[_mem.tx_tone]

        rxpol = None
        if _mem.rx_tone_off:
            rxmode = None
            rxval = None
        elif _mem.rx_tone_digital:
            rxmode = "DTCS"
            rxval = chirp_common.ALL_DTCS_CODES[int(_mem.rx_tone)]
        else:
            rxmode = "Tone"
            rxval = TONES[_mem.rx_tone]

        txspec = (txmode, txval, txpol)
        rxspec = (rxmode, rxval, rxpol)
        chirp_common.split_tone_decode(memory, txspec, rxspec)

        # RF Power
        if _mem.power_rf == 3:
            memory.power = self._power_high
        elif _mem.power_rf == 2:
            memory.power = self._power_low2
        else:
            memory.power = self._power_low1

        memory.extra = RadioSettingGroup('extra', 'Extra')

        # Mask (Don't display) channel on radio.
        # This is referred as Inhibit in the CS-F500 software.
        _disp_inhibit = RadioSetting(
            "disp_inhibit", "Mask",
            RadioSettingValueList(
                YES_CHOICE, current_index=int(_mem.disp_inhibit)))
        _disp_inhibit.set_doc(
            "Prevent the radio from displaying the channel.  This "
            "prevents use of the channel but CHIRP will still download "
            "the channel information.")
        memory.extra.append(_disp_inhibit)

        _compander_on = RadioSetting(
            "compander_on", "Compander",
            RadioSettingValueList(
                ON_CHOICE, current_index=int(
                    _mem.compander_on)))
        _compander_on.set_doc("Compander On")
        memory.extra.append(_compander_on)

        _tot_on = RadioSetting(
            "tot_on", "TOT",
            RadioSettingValueList(
                ON_CHOICE, current_index=int(_mem.tot_on)))
        _tot_on.set_doc("Time Out Timer On")
        memory.extra.append(_tot_on)

        _auto_reset_timer = RadioSetting(
            "auto_reset_timer", "Auto Reset Timer",
            RadioSettingValueList(
                RESET_TIMERS_VALUES, current_index=int(
                    _mem.auto_reset)))
        _auto_reset_timer.set_doc("Auto Reset Timer Selection")
        memory.extra.append(_auto_reset_timer)

        _lockout_bits = (_mem.lockout_busy << 1) | _mem.lockout_repeater
        _lockout = RadioSetting(
            "lockout", "Lockout",
            RadioSettingValueList(
                LOCKOUT_VALUES, current_index=int(_lockout_bits)))
        _lockout.set_doc("Transmit Lockout")
        memory.extra.append(_lockout)

        _log_in_out_bits = (_mem.log_in << 2) + _mem.log_out
        _log_in_out_index = 0
        if _log_in_out_bits == 0x03:
            _log_in_out_index = 2
        elif _log_in_out_bits == 0x0C:
            _log_in_out_index = 1
        elif _log_in_out_bits == 0x0F:
            _log_in_out_index = 3
        _log_in_out = RadioSetting(
            "log_in_out", "Log",
            RadioSettingValueList(
                LOG_IN_OUT_VALUES, current_index=int(_log_in_out_index)))
        _log_in_out.set_doc("Log in and Log out")
        memory.extra.append(_log_in_out)

        # Any memory with a frequency of 0 MHz is treated as empty.
        #
        # When CS-F500 inserts a blank memory, it sets Tx and Rx frequency
        # to 0 MHz and leaves the disp_inhibit bit unset.  When the
        # disp_inhibit bit is set, the memory is displayed in CS-F500
        # with an "i" in the inhibit column.
        #
        # The radio does not display inhibited (masked) memory
        # channels.
        if memory.freq == 0:
            memory.empty = True

        return memory

    def set_memory(self, memory):
        """Set the memory object @memory"""
        # Store details about a high-level memory to the memory map
        # This is called when a user edits a memory in the UI

        # Get a low-level memory object mapped to the image
        _mem = self._memobj.memory[memory.number - 1]

        rf = self.get_features()

        # Frequency: Low level format is frequency in Hz
        _mem.freq_rx = memory.freq
        _mem.tx_inhibit = 0
        if memory.duplex == "off":
            _mem.freq_tx = memory.freq
            _mem.tx_inhibit = 1
        elif memory.duplex == "split":
            _mem.freq_tx = memory.offset
        elif memory.duplex == "-":
            _mem.freq_tx = _mem.freq_rx - memory.offset
        elif memory.duplex == "+":
            _mem.freq_tx = _mem.freq_rx + memory.offset
        else:
            _mem.freq_tx = memory.freq

        # Modes
        if memory.mode == "NFM":
            _mem.narrow = 1
        else:
            _mem.narrow = 0

        # Name (alpha tag)
        _mem.name = str(memory.name).ljust(rf.valid_name_length)[
            :rf.valid_name_length]
        _mem.text_on = 1

        # Tone
        txspec, rxspec = chirp_common.split_tone_encode(memory)
        txmode, txval, txpol = txspec
        rxmode, rxval, rxpol = rxspec

        _mem.tx_tone_off = 1
        _mem.tx_tone_digital = 0
        _mem.tx_tone = 0
        if txmode == "Tone":
            _mem.tx_tone_off = 0
            _mem.tx_tone = TONES.index(txval)
        elif txmode == "DTCS":
            _mem.tx_tone_off = 0
            _mem.tx_tone_digital = 1
            _mem.tx_tone = chirp_common.ALL_DTCS_CODES.index(txval)

        _mem.rx_tone_off = 1
        _mem.rx_tone_digital = 0
        _mem.rx_tone = 0
        if rxmode == "Tone":
            _mem.rx_tone_off = 0
            _mem.rx_tone = TONES.index(rxval)
        elif rxmode == "DTCS":
            _mem.rx_tone_off = 0
            _mem.rx_tone_digital = 1
            _mem.rx_tone = chirp_common.ALL_DTCS_CODES.index(rxval)

        # RF Power
        if memory.power == self._power_high:
            _mem.power_rf = 3
        elif memory.power == self._power_low2:
            _mem.power_rf = 2
        else:
            _mem.power_rf = 1

        _mem.disp_inhibit = 0
        _mem.compander_on = 0
        _mem.tot_on = 0
        _mem.auto_reset = 0
        _mem.lockout_busy = 0
        _mem.lockout_repeater = 0
        _mem.log_in = 0
        _mem.log_out = 0

        if memory.extra:
            _mem.disp_inhibit = memory.extra['disp_inhibit'].value
            _mem.compander_on = memory.extra['compander_on'].value
            _mem.tot_on = memory.extra['tot_on'].value
            _mem.auto_reset = memory.extra['auto_reset_timer'].value

            _lockout = int(memory.extra['lockout'].value)
            _mem.lockout_repeater = _lockout & 1
            _mem.lockout_busy = (_lockout >> 1) & 1

            # There are two bits each for log in and log out/off.
            # 0 is off and 3 is on.  1 and 2 remain a mystery
            _log_in_out = int(memory.extra['log_in_out'].value)
            _mem.log_in = 0
            _mem.log_out = 0
            if _log_in_out == 1:
                _mem.log_in = 3
            elif _log_in_out == 2:
                _mem.log_out = 3
            elif _log_in_out == 3:
                _mem.log_in = 3
                _mem.log_out = 3

        if memory.empty:
            _mem.freq_rx = 0
            _mem.freq_tx = 0
            _mem.narrow = 0
            _mem.name = "          "
            _mem.text_on = 0
            _mem.tx_tone_off = 1
            _mem.tx_tone_digital = 0
            _mem.tx_tone = 0
            _mem.rx_tone_off = 1
            _mem.rx_tone_digital = 0
            _mem.rx_tone = 0

# Radio Settings

    # Build the "DTMF" menu.
    # TODO: This menu is not even started.

    # Build the "Continuous Tone" menu.
    # TODO: This menu is not complete.
    def _get_continuous_tone_group(self):
        _group = RadioSettingGroup(
            "group_continuous_tone", "Continuous Tone")
        # 9 custom CTCSS pairs
        _group.append(RadioSetting(
            "tone_burst_phase_on", "Tone Burst",
            RadioSettingValueList(
                TONE_BURST_VALUES, current_index=int(
                    self._memobj.tone_burst_phase_on))))
        # "CTCSS Reverse Burst Timer"
        #  "User CTCSS Freq (Hz)"
        _group.append(RadioSetting(
            "tx_dtcs_inverse", "TX DTCS Inverse",
            RadioSettingValueList(
                INVERSE_CHOICE, current_index=int(
                    self._memobj.tx_dtcs_inverse))))
        _group.append(RadioSetting(
            "rx_dtcs_inverse", "RX DTCS Inverse",
            RadioSettingValueList(
                INVERSE_CHOICE, current_index=int(
                    self._memobj.rx_dtcs_inverse))))
        _group.append(RadioSetting(
            "rx_af_filter_low_cut_on", "RX AF Filter",
            RadioSettingValueList(
                AUDIO_FILTER_RX_OPTIONS, current_index=int(
                    self._memobj.rx_af_filter_low_cut_on))))
        return _group

    # Build the "Scan" menu.
    # TODO: This menu is barely started
    def _get_scan_group(self):
        _group = RadioSettingGroup("group_scan", "Scan")
        # 10 scan lists
        # timer.scan_stop   "Stop"
        # timer.scan_resume "Resume"
        # timer.scan_fast   "Fast Scan
        # timer.scan_slow   "Slow Scan"
        _group.append(RadioSetting(
            "poweron_scan_on", "Power on Scan",
            RadioSettingValueList(
                ON_OFF_CHOICE, current_index=int(
                    self._memobj.poweron_scan_on))))
        return _group

    # Build the "2TONE" menu
    # TODO: This menu is not even started.

    # Build the "Conventional Key" menu.
    def _get_conv_key_group(self):
        _group = RadioSettingGroup("conv_keys", "Conventional Keys")
        _group.append(RadioSetting(
            "conv_key_lup", "L-Up",
            RadioSettingValueList(
                KEY_CONV, current_index=kc_kf(
                    self._memobj.conv_key.lup))))
        _group.append(RadioSetting(
            "conv_key_ldown", "L-Down",
            RadioSettingValueList(
                KEY_CONV, current_index=kc_kf(
                    self._memobj.conv_key.ldown))))
        _group.append(RadioSetting(
            "conv_key_rup", "R-Up",
            RadioSettingValueList(
                KEY_CONV, current_index=kc_kf(
                    self._memobj.conv_key.rup))))
        _group.append(RadioSetting(
            "conv_key_rdown", "R-Down",
            RadioSettingValueList(
                KEY_CONV, current_index=kc_kf(
                    self._memobj.conv_key.rdown))))
        _group.append(RadioSetting(
            "conv_key_p0", "P0",
            RadioSettingValueList(
                KEY_CONV, current_index=kc_kf(
                    self._memobj.conv_key.p0))))
        _group.append(RadioSetting(
            "conv_key_p1", "P1",
            RadioSettingValueList(
                KEY_CONV, current_index=kc_kf(
                    self._memobj.conv_key.p1))))
        _group.append(RadioSetting(
            "conv_key_p2", "P2",
            RadioSettingValueList(
                KEY_CONV, current_index=kc_kf(
                    self._memobj.conv_key.p2))))
        _group.append(RadioSetting(
            "conv_key_p3", "P3",
            RadioSettingValueList(
                KEY_CONV, current_index=kc_kf(
                    self._memobj.conv_key.p3))))
        _group.append(RadioSetting(
            "conv_key_p4", "P4",
            RadioSettingValueList(
                KEY_CONV, current_index=kc_kf(
                    self._memobj.conv_key.p4))))
        _group.append(RadioSetting(
            "conv_key_opf0", "OPF0",
            RadioSettingValueList(
                KEY_CONV, current_index=kc_kf(
                    self._memobj.conv_key.opf0))))
        _group.append(RadioSetting(
            "conv_key_opf1", "OPF1",
            RadioSettingValueList(
                KEY_CONV, current_index=kc_kf(
                    self._memobj.conv_key.opf1))))
        _group.append(RadioSetting(
            "conv_key_opf2", "OPF2",
            RadioSettingValueList(
                KEY_CONV, current_index=kc_kf(
                    self._memobj.conv_key.opf2))))
        return _group

    # Build the "SmarTrunk Key" menu.
    def _get_trunk_key_group(self):
        _group = RadioSettingGroup("trunk_keys", "SmarTrunk Keys")
        _group.append(RadioSetting(
            "trunk_key_lup", "L-Up",
            RadioSettingValueList(
                KEY_TRUNK, current_index=kt_kf(
                    self._memobj.trunk_key.lup))))
        _group.append(RadioSetting(
            "trunk_key_ldown", "L-Down",
            RadioSettingValueList(
                KEY_TRUNK, current_index=kt_kf(
                    self._memobj.trunk_key.ldown))))
        _group.append(RadioSetting(
            "trunk_key_rup", "R-Up",
            RadioSettingValueList(
                KEY_TRUNK, current_index=kt_kf(
                    self._memobj.trunk_key.rup))))
        _group.append(RadioSetting(
            "trunk_key_rdown", "R-Down",
            RadioSettingValueList(
                KEY_TRUNK, current_index=kt_kf(
                    self._memobj.trunk_key.rdown))))
        _group.append(RadioSetting(
            "trunk_key_p0", "P0",
            RadioSettingValueList(
                KEY_TRUNK, current_index=kt_kf(
                    self._memobj.trunk_key.p0))))
        _group.append(RadioSetting(
            "trunk_key_p1", "P1",
            RadioSettingValueList(
                KEY_TRUNK, current_index=kt_kf(
                    self._memobj.trunk_key.p1))))
        _group.append(RadioSetting(
            "trunk_key_p2", "P2",
            RadioSettingValueList(
                KEY_TRUNK, current_index=kt_kf(
                    self._memobj.trunk_key.p2))))
        _group.append(RadioSetting(
            "trunk_key_p3", "P3",
            RadioSettingValueList(
                KEY_TRUNK, current_index=kt_kf(
                    self._memobj.trunk_key.p3))))
        _group.append(RadioSetting(
            "trunk_key_p4", "P4",
            RadioSettingValueList(
                KEY_TRUNK, current_index=kt_kf(
                    self._memobj.trunk_key.p4))))
        _group.append(RadioSetting(
            "trunk_key_opf0", "OPF0",
            RadioSettingValueList(
                KEY_TRUNK, current_index=kt_kf(
                    self._memobj.trunk_key.opf0))))
        _group.append(RadioSetting(
            "trunk_key_opf1", "OPF1",
            RadioSettingValueList(
                KEY_TRUNK, current_index=kt_kf(
                    self._memobj.trunk_key.opf1))))
        _group.append(RadioSetting(
            "trunk_key_opf2", "OPF2",
            RadioSettingValueList(
                KEY_TRUNK, current_index=kt_kf(
                    self._memobj.trunk_key.opf2))))
        return _group

    # Build the "Display" menu.
    def _get_display_group(self):
        _freq_lo = self._memobj.beep_frequency_low / 10.0
        _freq_hi = self._memobj.beep_frequency_high / 10.0
        _psw = int(not self._memobj.poweron_selection_off)
        _isw = int(self._memobj.ignition_switch_on)
        _pria = int(self._memobj.poweron_priority_a)

        _group = RadioSettingGroup("group_display", "Display")
        _group.append(RadioSetting(
            "beep_frequency_low", "Beep Low Freq",
            RadioSettingValueFloat(
                400.0, 3000.0, _freq_lo, precision=0)))
        _group.append(RadioSetting(
            "beep_frequency_high", "Beep High Freq",
            RadioSettingValueFloat(
                400.0, 3000.0, _freq_hi, precision=0)))
        _group.append(RadioSetting(
            "opening_text", "Opening Text",
            RadioSettingValueString(
                0, 10, str(self._memobj.opening_text).rstrip(), False)))
        _group.append(RadioSetting(
            "rf_power_selection", "RF Power",
            RadioSettingValueList(
                RF_POWER_SELECTION_OPTIONS, current_index=int(
                    self._memobj.rf_power_selection))))
        _group.append(RadioSetting(
            "poweron_selection_off", "Power Switch Selection",
            RadioSettingValueList(APPEAR_CHOICE, current_index=_psw)))
        _group.append(RadioSetting(
            "ignition_switch_on", "Ignition Switch",
            RadioSettingValueList(APPEAR_CHOICE, current_index=_isw)))
        _group.append(RadioSetting(
            "ignition_switch_delay", "Ignition Switch Delay Timer",
            RadioSettingValueInteger(
                0, 240, self._memobj.timer.ignition_switch_delay)))
        _group.append(RadioSetting(
            "poweron_priority_a", "Power On Priority A Channel",
            RadioSettingValueList(ENABLE_CHOICE, current_index=_pria)))
        _group.append(RadioSetting(
            "hookoff_priority_a", "Hook Off Priority A Channel",
            RadioSettingValueList(
                ENABLE_CHOICE, current_index=int(
                    self._memobj.hookoff_priority_a))))
        _group.append(RadioSetting(
            "hookoff_monitor_on", "Hook Off Monitor",
            RadioSettingValueList(
                ON_OFF_CHOICE, current_index=int(
                    self._memobj.hookoff_monitor_on))))
        _group.append(RadioSetting(
            "hookon_scan_on", "Hook on Scan",
            RadioSettingValueList(
                ON_OFF_CHOICE, current_index=int(
                    self._memobj.hookon_scan_on))))
        return _group

    # Build the "Set Mode" menu.
    def _get_set_mode_group(self):
        _hrn = int(not self._memobj.horn_off)

        _group = RadioSettingGroup("set_mode_group", "Set Mode")
        _group.append(RadioSetting(
            "backlight", "Backlight",
            RadioSettingValueList(
                BACKLIGHT_VALUES, current_index=int(
                    self._memobj.backlight))))
        _group.append(RadioSetting(
            "backlight_set_enable", "Enable Backlight",
            RadioSettingValueList(
                NO_YES_CHOICE, current_index=int(
                    self._memobj.backlight_set_enable))))
        _group.append(RadioSetting(
            "beep_on", "Beep",
            RadioSettingValueList(
                ON_OFF_CHOICE, current_index=int(
                    self._memobj.beep_on))))
        _group.append(RadioSetting(
            "beep_set_enable", "Enable Beep",
            RadioSettingValueList(
                NO_YES_CHOICE, current_index=int(
                    self._memobj.beep_set_enable))))
        _group.append(RadioSetting(
            "beep_level", "Beep Level",
            RadioSettingValueList(
                BEEP_LEVEL_VALUES, current_index=int(
                    self._memobj.beep_level))))
        _group.append(RadioSetting(
            "beep_level_set_enable", "Enable Beep Level",
            RadioSettingValueList(
                NO_YES_CHOICE, current_index=int(
                    self._memobj.beep_level_set_enable))))
        _group.append(RadioSetting(
            "squelch_level", "Squelch Level",
            RadioSettingValueInteger(
                0, 255, self._memobj.squelch_level)))
        _group.append(RadioSetting(
            "squelch_level_set_enable", "Enable Squelch Level",
            RadioSettingValueList(
                NO_YES_CHOICE, current_index=int(
                    self._memobj.squelch_level_set_enable))))
        _group.append(RadioSetting(
            "volume_min_level", "Volume Minimum Level",
            RadioSettingValueInteger(
                0, 255, self._memobj.volume_min_level)))
        _group.append(RadioSetting(
            "volume_min_level_set_enable", "Enable Volume Minimum Level",
            RadioSettingValueList(
                NO_YES_CHOICE, current_index=int(
                    self._memobj.volume_min_level_set_enable))))
        _group.append(RadioSetting(
            "audio_filter", "Audio Filter",
            RadioSettingValueList(
                AUDIO_FILTER_VALUES, current_index=int(
                    self._memobj.audio_filter))))
        _group.append(RadioSetting(
            "audio_filter_set_enable", "Enable Audio Filter",
            RadioSettingValueList(
                NO_YES_CHOICE, current_index=int(
                    self._memobj.audio_filter_set_enable))))
        _group.append(RadioSetting(
            "mic_gain", "Mic Gain",
            RadioSettingValueList(
                MIC_GAIN_CHOICE, current_index=mc_mf(
                    self._memobj.mic_gain))))
        _group.append(RadioSetting(
            "mic_gain_set_enable", "Enable Mic Gain",
            RadioSettingValueList(
                NO_YES_CHOICE, current_index=int(
                    self._memobj.mic_gain_set_enable))))
        _group.append(RadioSetting(
            "horn_off", "Horn",
            RadioSettingValueList(
                ON_CHOICE, current_index=_hrn)))
        _group.append(RadioSetting(
            "horn_set_enable", "Enable Horn",
            RadioSettingValueList(
                NO_YES_CHOICE, current_index=int(
                    self._memobj.horn_set_enable))))
        return _group

    # Build the "Common" menu.
    def _get_common_group(self):
        _timer_eptt = self._memobj.timer.eptt_delay / 10.0
        _timer_exo = self._memobj.timer.exo_delay / 10.0
        _timer_es_on = self._memobj.timer.emergency_switch_on / 10.0
        _timer_es_off = self._memobj.timer.emergency_switch_off / 10.0

        _group = RadioSettingGroup("common_group", "Common")
        _group.append(RadioSetting(
            "clone_comment", "Clone Comment",
            RadioSettingValueString(
                0, 32, str(self._memobj.clone_comment).rstrip(), False)))
        _group.append(RadioSetting(
            "password_on", "Power-on Password",
            RadioSettingValueList(
                ON_OFF_CHOICE, current_index=int(
                    self._memobj.password_on))))
        _group.append(RadioSetting(
            "data_out_on", "Transceiver Data Out",
            RadioSettingValueList(
                ON_OFF_CHOICE, current_index=int(
                    self._memobj.data_out_on))))
        _group.append(RadioSetting(
            "timer_auto_reset_a", "Auto Reset Timer A",
            RadioSettingValueInteger(
                0, 255, self._memobj.timer.auto_reset_a)))
        _group.append(RadioSetting(
            "timer_auto_reset_b", "Auto Reset Timer B",
            RadioSettingValueInteger(
                0, 255, self._memobj.timer.auto_reset_b)))
        _group.append(RadioSetting(
            "timer_time_out", "Time Out Timer",
            RadioSettingValueInteger(
                0, 255, self._memobj.timer.time_out)))
        _group.append(RadioSetting(
            "timer_penalty", "Penalty Timer",
            RadioSettingValueInteger(
                0, 255, self._memobj.timer.penalty)))
        _group.append(RadioSetting(
            "dtmf_id_out_on", "DTMF ID Out",
            RadioSettingValueList(
                ON_OFF_CHOICE, current_index=int(
                    self._memobj.dtmf_id_out_on))))
        _group.append(RadioSetting(
            "time_out_timer_beep_on", "Time Out Timer Beep",
            RadioSettingValueList(
                ON_OFF_CHOICE, current_index=int(
                    self._memobj.time_out_timer_beep_on))))
        _group.append(RadioSetting(
            "timer_lockout_penalty", "Lockout Penalty Timer",
            RadioSettingValueInteger(
                0, 255, self._memobj.timer.lockout_penalty)))
        _group.append(RadioSetting(
            "timer_eptt_delay", "EPTT Delay Timer",
            RadioSettingValueFloat(0, 25.5, _timer_eptt, precision=1)))
        _group.append(RadioSetting(
            "tone_mute_eptt_on", "Tone Mute EPTT",
            RadioSettingValueList(
                ON_OFF_CHOICE, current_index=int(
                    self._memobj.tone_mute_eptt_on))))
        _group.append(RadioSetting(
            "exo_on", "EXO", RadioSettingValueList(
                ON_OFF_CHOICE, current_index=int(self._memobj.exo_on))))
        _group.append(RadioSetting(
            "timer_exo_delay", "EXO Delay Timer",
            RadioSettingValueFloat(0, 25.5, _timer_exo, precision=1)))
        _group.append(RadioSetting(
            "timer_horn", "EXO/Horn Timer",
            RadioSettingValueInteger(0, 255, self._memobj.timer.horn)))
        _group.append(RadioSetting(
            "scrambler_non_rolling", "Scrambler Type",
            RadioSettingValueList(
                SCRAMBLER_TYPE_VALUES, current_index=int(
                    self._memobj.scrambler_non_rolling))))
        _group.append(RadioSetting(
            "scrambler_group_code", "Scrambler Group Code",
            RadioSettingValueList(
                SCRAMBLER_GCODE_VALUES, current_index=int(
                    self._memobj.scrambler_group_code))))
        _group.append(RadioSetting(
            "scrambler_syn_capture", "Scrambler Synchronous Capture",
            RadioSettingValueList(
                SYN_CAPTURE_CHOICE, current_index=sc_sf(
                    self._memobj.scrambler_syn_capture))))
        _group.append(RadioSetting(
            "scrambler_tone_start_timing", "Scrambler Tone Start Timing",
            RadioSettingValueList(
                TONE_STIMING_VALUES, current_index=int(
                    self._memobj.scrambler_tone_start_timing))))
        _group.append(RadioSetting(
            "timer_emergency_switch_on", "Emergency Switch On Timer",
            RadioSettingValueFloat(0, 25.5, _timer_es_on, precision=1)))
        _group.append(RadioSetting(
            "timer_emergency_switch_off", "Emergency Switch Off Timer",
            RadioSettingValueFloat(0, 25.5, _timer_es_off, precision=1)))
        _group.append(RadioSetting(
            "emergency_start_repeat", "Emergency Start/Repeat",
            RadioSettingValueInteger(
                0, 255, self._memobj.timer.emergency_start_repeat)))
        return _group

    # Build the top menu.
    def get_settings(self):
        """Returns a RadioSettingGroup containing one or more
        RadioSettingGroup or RadioSetting objects. These represent general
        setting knobs and dials that can be adjusted on the radio."""
        return list(
            RadioSettingGroup(
                "top", "LMR",
                self._get_continuous_tone_group(),
                self._get_scan_group(),
                self._get_conv_key_group(),
                self._get_trunk_key_group(),
                self._get_display_group(),
                self._get_set_mode_group(),
                self._get_common_group()))

    def set_settings(self, settings):
        """Accepts the top-level RadioSettingGroup returned from get_settings()
        and adjusts the values in the radio accordingly. This function expects
        the entire RadioSettingGroup hierarchy returned from get_settings()."""
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)

            # "Continuous Tone" items
            elif element.get_name() == "tone_burst_phase_on":
                self._memobj.tone_burst_phase_on = element.value
            elif element.get_name() == "tx_dtcs_inverse":
                self._memobj.tx_dtcs_inverse = element.value
            elif element.get_name() == "rx_dtcs_inverse":
                self._memobj.rx_dtcs_inverse = element.value
            elif element.get_name() == "rx_af_filter_low_cut_on":
                self._memobj.rx_af_filter_low_cut_on = element.value

            # "Scan" items
            elif element.get_name() == "poweron_scan_on":
                self._memobj.poweron_scan_on = element.value

            # "Conventional Key" items
            elif element.get_name() == "conv_key_lup":
                self._memobj.conv_key.lup = kf_kc(element.value)
            elif element.get_name() == "conv_key_ldown":
                self._memobj.conv_key.ldown = kf_kc(element.value)
            elif element.get_name() == "conv_key_rup":
                self._memobj.conv_key.rup = kf_kc(element.value)
            elif element.get_name() == "conv_key_rdown":
                self._memobj.conv_key.rdown = kf_kc(element.value)
            elif element.get_name() == "conv_key_p0":
                self._memobj.conv_key.p0 = kf_kc(element.value)
            elif element.get_name() == "conv_key_p1":
                self._memobj.conv_key.p1 = kf_kc(element.value)
            elif element.get_name() == "conv_key_p2":
                self._memobj.conv_key.p2 = kf_kc(element.value)
            elif element.get_name() == "conv_key_p3":
                self._memobj.conv_key.p3 = kf_kc(element.value)
            elif element.get_name() == "conv_key_p4":
                self._memobj.conv_key.p4 = kf_kc(element.value)
            elif element.get_name() == "conv_key_opf0":
                self._memobj.conv_key.opf0 = kf_kc(element.value)
            elif element.get_name() == "conv_key_opf1":
                self._memobj.conv_key.opf1 = kf_kc(element.value)
            elif element.get_name() == "conv_key_opf2":
                self._memobj.conv_key.opf2 = kf_kc(element.value)
            elif element.get_name() == "trunk_key_lup":
                self._memobj.trunk_key.lup = kf_kt(element.value)

            # "SmarTrunk Key" items
            elif element.get_name() == "trunk_key_ldown":
                self._memobj.trunk_key.ldown = kf_kt(element.value)
            elif element.get_name() == "trunk_key_rup":
                self._memobj.trunk_key.rup = kf_kt(element.value)
            elif element.get_name() == "trunk_key_rdown":
                self._memobj.trunk_key.rdown = kf_kt(element.value)
            elif element.get_name() == "trunk_key_p0":
                self._memobj.trunk_key.p0 = kf_kt(element.value)
            elif element.get_name() == "trunk_key_p1":
                self._memobj.trunk_key.p1 = kf_kt(element.value)
            elif element.get_name() == "trunk_key_p2":
                self._memobj.trunk_key.p2 = kf_kt(element.value)
            elif element.get_name() == "trunk_key_p3":
                self._memobj.trunk_key.p3 = kf_kt(element.value)
            elif element.get_name() == "trunk_key_p4":
                self._memobj.trunk_key.p4 = kf_kt(element.value)
            elif element.get_name() == "trunk_key_opf0":
                self._memobj.trunk_key.opf0 = kf_kt(element.value)
            elif element.get_name() == "trunk_key_opf1":
                self._memobj.trunk_key.opf1 = kf_kt(element.value)
            elif element.get_name() == "trunk_key_opf2":
                self._memobj.trunk_key.opf2 = kf_kt(element.value)

            # "Display Mode" items
            elif element.get_name() == "beep_frequency_low":
                self._memobj.beep_frequency_low = element.value * 10.0
            elif element.get_name() == "beep_frequency_high":
                self._memobj.beep_frequency_high = element.value * 10.0
            elif element.get_name() == "opening_text":
                self._memobj.opening_text = str(
                    element.value).ljust(10)[:10]
            elif element.get_name() == "rf_power_selection":
                self._memobj.rf_power_selection = element.value
            elif element.get_name() == "poweron_selection_off":
                self._memobj.poweron_selection_off = not element.value
            elif element.get_name() == "ignition_switch_on":
                self._memobj.ignition_switch_on = element.value
            elif element.get_name() == "ignition_switch_delay":
                self._memobj.timer.ignition_switch_delay = element.value
            elif element.get_name() == "poweron_priority_a":
                self._memobj.poweron_priority_a = element.value
            elif element.get_name() == "hookoff_priority_a":
                self._memobj.hookoff_priority_a = element.value
            elif element.get_name() == "hookoff_monitor_on":
                self._memobj.hookoff_monitor_on = element.value
            elif element.get_name() == "hookon_scan_on":
                self._memobj.hookon_scan_on = element.value

            # "Set Mode" items
            elif element.get_name() == "backlight":
                self._memobj.backlight = element.value
            elif element.get_name() == "backlight_set_enable":
                self._memobj.backlight_set_enable = element.value
            elif element.get_name() == "beep_on":
                self._memobj.beep_on = element.value
            elif element.get_name() == "beep_set_enable":
                self._memobj.beep_set_enable = element.value
            elif element.get_name() == "beep_level":
                self._memobj.beep_level = element.value
            elif element.get_name() == "beep_level_set_enable":
                self._memobj.beep_level_set_enable = element.value
            elif element.get_name() == "squelch_level":
                self._memobj.squelch_level = element.value
            elif element.get_name() == "squelch_level_set_enable":
                self._memobj.squelch_level_set_enable = element.value
            elif element.get_name() == "volume_min_level":
                self._memobj.volume_min_level = element.value
            elif element.get_name() == "volume_min_level_set_enable":
                self._memobj.volume_min_level_set_enable = element.value
            elif element.get_name() == "audio_filter":
                self._memobj.audio_filter = element.value
            elif element.get_name() == "audio_filter_set_enable":
                self._memobj.audio_filter_set_enable = element.value
            elif element.get_name() == "mic_gain":
                self._memobj.mic_gain = mf_mc(element.value)
            elif element.get_name() == "mic_gain_set_enable":
                self._memobj.mic_gain_set_enable = element.value
            elif element.get_name() == "horn_off":
                self._memobj.horn_off = not element.value
            elif element.get_name() == "horn_set_enable":
                self._memobj.horn_set_enable = element.value

            # "Common" items
            elif element.get_name() == "clone_comment":
                self._memobj.clone_comment = str(
                    element.value).ljust(32)[:32]
            elif element.get_name() == "password_on":
                self._memobj.password_on = element.value
            elif element.get_name() == "data_out_on":
                self._memobj.data_out_on = element.value
            elif element.get_name() == "timer_auto_reset_a":
                self._memobj.timer.auto_reset_a = element.value
            elif element.get_name() == "timer_auto_reset_b":
                self._memobj.timer.auto_reset_b = element.value
            elif element.get_name() == "timer_time_out":
                self._memobj.timer.time_out = element.value
            elif element.get_name() == "timer_penalty":
                self._memobj.timer.penalty = element.value
            elif element.get_name() == "dtmf_id_out_on":
                self._memobj.dtmf_id_out_on = element.value
            elif element.get_name() == "time_out_timer_beep_on":
                self._memobj.time_out_timer_beep_on = element.value
            elif element.get_name() == "timer_lockout_penalty":
                self._memobj.timer.lockout_penalty = element.value
            elif element.get_name() == "timer_eptt_delay":
                self._memobj.timer.eptt_delay = element.value * 10.0
            elif element.get_name() == "tone_mute_eptt_on":
                self._memobj.tone_mute_eptt_on = element.value
            elif element.get_name() == "exo_on":
                self._memobj.tone_mute_eptt_on = element.value
            elif element.get_name() == "timer_exo_delay":
                self._memobj.timer.exo_delay = element.value * 10.0
            elif element.get_name() == "timer_horn":
                self._memobj.timer.horn = element.value
            elif element.get_name() == "scrambler_non_rolling":
                self._memobj.scrambler_non_rolling = element.value
            elif element.get_name() == "scrambler_group_code":
                self._memobj.scrambler_group_code = element.value
            elif element.get_name() == "scrambler_syn_capture":
                self._memobj.scrambler_syn_capture = sf_sc(element.value)
            elif element.get_name() == "scrambler_tone_start_timing":
                self._memobj.scrambler_tone_start_timing = element.value
            elif element.get_name() == "timer_emergency_switch_on":
                self._memobj.timer.emergency_switch_on = element.value * 10.0
            elif element.get_name() == "timer_emergency_switch_off":
                self._memobj.timer.emergency_switch_off = element.value * 10.0
            elif element.get_name() == "emergency_start_repeat":
                self._memobj.timer.emergency_start_repeat = element.value
            else:
                LOG.warning(
                    'Attempt to set _memobj.%s to "%s"'
                    % (element.get_name(), element.value))



import sys
from unittest import mock

from chirp import chirp_common
import lark
from tests.unit import base

sys.modules['wx'] = wx = mock.MagicMock()
sys.modules['wx.adv'] = wx = mock.MagicMock()

from chirp.wxui import memquery  # noqa


class TestMemquery(base.BaseTest):
    def _get_sample_memories(self):
        mems = []
        freqs = (118, 145, 146, 440, 800)
        for i, freq in enumerate(freqs):
            mem = chirp_common.Memory(1 + i, name="mem%s" % freq)
            mem.freq = freq * 1000000
            mem.mode = "FM"
            mems.append(mem)
        return mems

    def test_query_parse(self):
        query = ('name="foo" OR name IN ["mem800", "baz"] OR '
                 '(mode="FM" AND freq<144,147.1>) OR '
                 'name~"7$"')
        parser = lark.Lark(memquery.LANG)
        transformer = memquery.Interpreter(self._get_sample_memories())
        filtered = transformer.transform(parser.parse(query)).children[0]
        self.assertEqual(3, len(filtered), filtered)
        self.assertEqual([145, 146, 800],
                         sorted([x.freq // 1000000 for x in filtered]))



# Copyright 2024 Tommy <tommy83033@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import struct
import logging
import math
import time
from enum import Enum

from chirp import memmap, chirp_common, bitwise, directory, errors
from chirp.settings import (
    RadioSetting,
    RadioSettingGroup,
    RadioSettingValueList,
    RadioSettingValueString,
    RadioSettings,
    RadioSettingValueBoolean,
    RadioSettingValueFloat)


LOG = logging.getLogger(__name__)

MEM_FORMAT = """

#seekto 0x0E;
struct{
    u8 modelnumber[32];
    u8 hardwareversion[2];
    u8 serialno[16];
    ul32 freqmin;
    ul32 freqmax;
    u8 rev_1;
    u8 saleterr;
    u8 radiomode;
    u8 rev[7];
} radioinfo;

#seekto 0x52;
struct{
    u8 dataver[2];
    u8 softver[4];
    u8 rev[4];
    }dataver;

#seekto 0x5C;
struct {
  u8 readpwd[8];
  u8 writepwd[8];
  u8 poweronpwd[8];
  u8 squelch:4,
     rev_1:4;
  u8 powersavingdelaytime:4,
     powersavingmode:4;
  u8 tk1_short;
  u8 tk1_long;
  u8 sk1_short;
  u8 sk1_long;
  u8 sk2_short;
  u8 sk2_long;
  u8 longtime;
  u8 chknoblock:1,
     sidekeylock:1,
     panellock:1,
     rev_2:3,
     autoormanuallock:1,
     keylockswitch:1;
  u8 wxscan:1,
     agc:1,
     rtcreport:1,
     displaybatterylevel:1,
     batterlevelreport:1,
     lowbattery:1,
     stunmode:2;
  u8 voiceprompts:1,
     touchtone:1,
     txpermittone:1,
     channelmode:1,
     screenlight:1,
     rogerbeep:1,
     rev_3:2;
  ul16 poweron_type_1:4,
       homepoweronzone_1:12;
  ul16 homepoweronch_1;
  u8 relaytaildelay;
  u8 radioid[10];
  u8 radioname[16];
  u8 micmain;
  u8 agcmain:4,
     agcthreshold:4;
  u8 scenemode;
  u8 autooff;
  u8 menuouttime;
  u8 voxthreshold:4,
     voxdelaytime:4;
  u8 rev_4:1,
     homeselect:3,
     homeindex:2,
     chdisplay:2;
  ul16 poweron_type_2:4,
     homepoweronzone_2:12;
  ul16 homepoweronch_2;
  ul16 homepoweronzone_3;
  ul16 homepoweronch_3;
  ul16 freqstep;
  u8 language;
  u8 backlightbrightness:4,
     backlighttime:4;
  u8 wxch;
  u8 calltone;
  u8 rev_5:5,
     homechtype_3:1,
     homechtype_2:1,
     homechtype_1:1;
  u8 spreadspectrummode;
  u8 salezone;
  u8 tailsoundeliminationsfre:7,
     tailsoundeliminationswitch:1;
  u8 singletone;
  u8 scanlist;
  u8 rev_6[6];
} settings;

#seekto 0xc0;
struct  {
    ul16 zonenum;
    ul16 zoneindex[64];
} zonedata;

#seekto 0x142;
struct {
    char name[14];
    ul16 chnum;
    ul16 chindex[16];
} zones[64];

#seekto 0x0D42;
struct  {
    ul16 chnum;
    ul16 chindex[1027];
} channeldata;

#seekto 0x154a;
struct  {
  char alias[14];
  u8 chmode:4,
     chpro:4;
  u8 fixedpower:1,
     fixedbandwidth:1,
     tailelimination:2,
     power:2,
     bandwidth:2;
  ul32 rxfreq;
  ul32 txfreq;
  u16 rxctcvaluetype:2,
      rxctctypecode:1,
      rev_1:1,
      rxctc:12;
  u8 rxsqlmode;
  u16 txctcvaluetype:2,
      txctctypecode:1,
      rev_2:1,
      txctc:12;
  u8 totpermissions:2,
     tottime:6;
  u8 vox:1,
     companding:1,
     scramble:1,
     offlineorreversal:2,
     rev_3:3;
  u8 voxthreshold:4,
     voxdelaytime:4;
  u8 autoscan:1,
     scanlist:7;
  u8 alarmlist;
  u8 dtmfsignalinglist:4,
     pttidtype:2,
     rev_4:2;
  ul16 dtmfcallid;
  u8 reserve[3];
} channels[1027];

#seekto 0xb5c2;
struct  {
    ul16 scannum;
    ul16 scanindex[16];
} scandata;

#seekto 0xb5e4;
struct {
   char name[14];
   u8 scantxch:4,
      scancondition:4;
   u8 hangtime:4,
      talkback:1,
      scanstatus:3;
   u8 chnum;
   ul16 specifych;
   ul16 PriorityCh1;
   ul16 PriorityCh2;
   u8 scanmode;
   ul16 chindex[100];
} scans[16];

#seekto 0xc3e4;
struct  {
    ul16 vfoscannum;
    ul16 vfoscanindex[3];
} vfoscandata;

#seekto 0xc3ec;
struct {
   u8 scantxch:4,
      scancondition:4;
   u8 hangtime:4,
      talkback:1,
      startcondition:1,
      rev_1:2;
   u8 scanmode;
   ul32 vhffreq_start;
   ul32 vhffreq_end;
   ul32 uhffreq_start;
   ul32 uhffreq_end;
   u8 rev;
} vfoscans[3];

#seekto 0xc444;
struct  {
    ul16 alarmnum;
    ul16 alarmindex[8];
} alarmdata;

#seekto 0xc456;
struct {
    char name[14];
    u8 alarmtype:4,
       alarmmode:4;
    ul16 jumpch;
    u8 localalarm:1,
       txbackground:1,
       ctcmode:2,
       rev_1:4;
    u8 alarmtime:4,
       alarmcycle:4;
    u8 mictime:4,
       txinterval:4;
    u8 alarmid[8];
    u8 alarmstatus;
    u8 rev_3[1];
    } alarms[8];

#seekto 0xc848;
struct  {
    ul16 dtmfnum;
    ul16 dtmfindex[4];
} dtmfdata;

#seekto 0xc852;
struct {
    u8 autoresettime:4,
       codedelaytime:4;
    u8 stunmode:2,
       showani:1,
       sidetone:1,
       pttidtype:2,
       rev_1:2;
    char callid[10];
    char stunid[10];
    char revive[10];
    char bot[16];
    char eot[16];
    char rev_2[8];
} dtmfcomm;

#seekto 0xc89a;
struct {
     char name[14];
     u8 codelen:4,
        signaling:4;
     u8 groupcode:4,
        intermediatecode:4;
     char fastcall1[16];
     char fastcall2[16];
     char fastcall3[16];
     char fastcall4[16];
     char fastcall5[16];
     char fastcall6[16];
     char fastcall7[16];
     char fastcall8[16];
     char fastcall9[16];
     char fastcall10[16];
     u8 rev_1:6,
        EncodingEnable:1,
        DecodingEnable:1;
     u8 dtmfstatus;
     u8 rev_2[12];
} dtmfs[4];
"""


class HandshakeStatuses(Enum):
    Normal = 0
    Wrong = 1
    PwdWrong = 3
    RadioWrong = 4


class MemoryRegions(Enum):
    """
    Defines the logical memory regions for this radio model.
    """
    radioHead = 2
    radioInfo = 3
    radioVer = 4
    settingData = 6
    zoneData = 7
    channelData = 8
    scanData = 11
    vfoScanData = 12
    alarmData = 13
    dTMFData = 15


MEMORY_REGIONS_RANGES = {
    MemoryRegions.radioHead: (0, 14),   # (Start addr, len)
    MemoryRegions.radioInfo: (14, 68),
    MemoryRegions.radioVer: (82, 10),
    MemoryRegions.settingData: (92, 100),
    MemoryRegions.zoneData: (192, 3202),
    MemoryRegions.channelData: (3394, 43136),
    MemoryRegions.scanData: (46530, 3618),
    MemoryRegions.vfoScanData: (50148, 68),
    MemoryRegions.alarmData: (50244, 258),
    MemoryRegions.dTMFData: (51272, 842)}

DTMFCHARSET = "0123456789ABCDabcd#*"
NAMECHARSET = chirp_common.CHARSET_ALPHANUMERIC + "-/;,._!? *#@$%&+=/<>~(){}]'"
SPECIAL_MEMORIES = {"VFOA": -2, "VFOB": -1}
BANDWIDEH_LIST = ["NFM", "FM"]
POWER_LEVELS = [
    chirp_common.PowerLevel("Low", watts=5),
    chirp_common.PowerLevel("High", watts=50)]
TIMEOUTTIMER_LIST = [{"name": "%ss" % (x * 5), "id": x}
                     for x in range(1, 64, 1)]
TOTPERMISSIONS_LIST = ["Always", "CTCSS/DCS Match",
                       "Channel Free", "Receive Only"]
SQUELCHLEVEL_LIST = ["AlwaysOpen", "1", "2", "3", "4", "5", "6", "7", "8", "9"]
AUTOSCAN_LIST = ["OFF", "Auto Scan System"]
OFFLINE_REVERSAL_LIST = ["OFF", "Freq Reversal", "Talk Around"]
SIDE_KEY_LIST = [
    {"name": "OFF", "id": 0},
    {"name": "TX Power", "id": 1},
    {"name": "Scan", "id": 2},
    {"name": "FM Radio", "id": 3},
    {"name": "Talkaround/Reversal", "id": 5},
    {"name": "Monitor", "id": 15},
    {"name": "Zone Plus", "id": 20},
    {"name": "Zone Minus", "id": 21},
    {"name": "Squelch", "id": 28},
    {"name": "Emergency Start", "id": 29},
    {"name": "Emergency Stop", "id": 30},
    {"name": "Optional DTMF Code", "id": 32},
    {"name": "Prog PTT", "id": 31}]
FREQ_STEP_List = [
    {"name": "2.5kHz", "id": 2500},
    {"name": "5kHz", "id": 5000},
    {"name": "6.25kHz", "id": 6250},
    {"name": "8.33kHz", "id": 8330},
    {"name": "10kHz", "id": 10000},
    {"name": "12.5kHz", "id": 12500},
    {"name": "25kHz", "id": 25000}]


def do_download(self):
    error_map = {
        HandshakeStatuses.RadioWrong: "Radio model mismatch",
        HandshakeStatuses.PwdWrong: "Radio is password protected"}
    try:
        handshake_result = handshake(self, self.pipe)
        if handshake_result == HandshakeStatuses.Normal:
            all_bytes = read_items(self, self.pipe)
            return memmap.MemoryMapBytes(bytes(all_bytes))
        raise errors.RadioError(error_map.get(
            handshake_result, "Unknown error communicating with radio"))
    except errors.RadioError:
        raise
    except Exception as e:
        LOG.error(f"download error: {e}")
        raise errors.RadioError("Unknown error communicating with radio")
    finally:
        exit_programming_mode(self)


def do_upload(self):
    error_map = {
        HandshakeStatuses.RadioWrong: "Radio model mismatch",
        HandshakeStatuses.PwdWrong: "Radio is password protected"}
    try:
        handshake_result = handshake(self, self.pipe)
        if handshake_result == HandshakeStatuses.Normal:
            write_items(self, self.pipe)
        else:
            raise errors.RadioError(error_map.get(
                handshake_result, "Unknown error communicating with radio"))
    except errors.RadioError:
        raise
    except Exception as e:
        LOG.error(f"upload error: {e}")
        raise errors.RadioError("Unknown error communicating with radio")
    finally:
        exit_programming_mode(self)


def handshake(self, serial):
    databytes = b""
    max_retries = 15
    retry_delay = 0.05
    for num in range(max_retries):
        flag, databytes = exchange_block_with_radio(
            get_handshake_bytes(self.MODEL + " "),
            serial, self.read_packet_len)
        time.sleep(retry_delay)
        if flag:
            break
    return validate_connection_handshake(self, databytes)


def read_items(self, serial):
    all_bytes = bytearray(self._memsize)
    status = chirp_common.Status()
    status.msg = "Cloning from radio"
    status.cur = 0
    status.max = self._memsize
    for item in MemoryRegions:
        try:
            item_bytes = get_read_current_packet_bytes(
                self, item.value, serial, status)
            if item_bytes:
                write_memory_region(all_bytes, item_bytes, item)
        except Exception as e:
            LOG.error(
                f"read item_data error: {item.name} error_msg: {e}")
            continue
    return all_bytes


def write_items(self, serial):
    status = chirp_common.Status()
    status.msg = "Uploading to radio"
    status.cur = 0
    status.max = self._memsize
    data_bytes = self.get_mmap()
    EXCLUDED_REGIONS = {MemoryRegions.radioHead,
                        MemoryRegions.radioInfo,
                        MemoryRegions.radioVer,
                        MemoryRegions.zoneData,
                        MemoryRegions.scanData,
                        MemoryRegions.alarmData}
    for item in MemoryRegions:
        if item in EXCLUDED_REGIONS:
            continue
        item_bytes = get_write_item_bytes(data_bytes, item)
        write_item_current_page_bytes(
            self, serial, item_bytes, item.value, status)


def validate_connection_handshake(self, dataByte: bytes):
    success_status = b"\x00\x01"
    pwd_faild_flag = 1
    if dataByte[14:16] != success_status:
        return HandshakeStatuses.Wrong
    if dataByte[20] == pwd_faild_flag:
        return HandshakeStatuses.PwdWrong
    radioType = dataByte[20: 20 + len(self.MODEL)]
    model_bytes = self.MODEL.encode("ascii")
    if radioType == model_bytes:
        return HandshakeStatuses.Normal
    else:
        return HandshakeStatuses.RadioWrong


def exchange_block_with_radio(
    send_byte: bytes, serial_conn, chunk_size=1050
) -> tuple[bool, bytes]:
    serial_conn.write(send_byte)
    new_bytes = serial_conn.read(chunk_size)
    if validate_packet(new_bytes):
        return True, new_bytes
    return False, new_bytes


def validate_packet(current_packet_Byte: bytes, chunk_size=1024):
    byteLen = len(current_packet_Byte)
    if byteLen <= 13:
        LOG.debug(f"Packet too short: {byteLen} bytes")
        return False
    """
    I noticed Retevis has another model
    that sends 250 bytes per packet,
    and the total packet count calculation is different.
    So I added this piece of code to handle
    that case and make it reusable later.
    """
    packet_count = 256 if chunk_size > 255 else chunk_size+1
    """
     I noticed that bytes [12:13] always equal len(current_page_byte) - 17.
     By adding this length check, we can filter out garbage data
     before it even gets to the CRC calculation - should give us a nice
     performance boost!
    """
    data_len = (current_packet_Byte[12] - 6
                if current_packet_Byte[12] > 6 else 0)
    data_len += current_packet_Byte[13] * packet_count
    if byteLen - 17 < data_len:
        LOG.debug(f"Packet too short: {byteLen} bytes")
        return False
    crcBytes = current_packet_Byte[-3:-1]
    crcStr = calculate_crc16(current_packet_Byte[:-3])
    return crcBytes == crcStr


def get_handshake_bytes(currentModel):
    handshake_flag = 0
    send_packet_index = 0
    send_packet_count = 1
    data_code = 0   # 0-write
    return get_send_packet_bytes(
        handshake_flag, send_packet_index, data_code, send_packet_count,
        currentModel.encode("ascii").ljust(41, b'\x00'))


def get_read_current_packet_bytes(self, item: int, serial, status):
    packet_count = 1
    packet_index = 0
    item_bytes = b""
    send_packet_index = 0
    send_packet_count = 1
    data_code = 2   # 0-write 2-read
    while packet_index < packet_count:
        flag, new_data_bytes = exchange_block_with_radio(
                get_send_packet_bytes(item, send_packet_index,
                                      data_code, send_packet_count,
                                      struct.pack("<H", packet_index)),
                serial, self.read_packet_len)
        if (not flag
           or (len(new_data_bytes) > 14 and new_data_bytes[14] != item)):
            raise errors.RadioError("radio reported failure")
        if packet_index == 0:
            packet_count, = struct.unpack("<H", new_data_bytes[18:20])
        status.cur += len(new_data_bytes)
        self.status_fn(status)
        item_bytes += new_data_bytes[20:-3]
        packet_index += 1
    return item_bytes


def write_item_current_page_bytes(self, serial, item_Bytes: bytes,
                                  item: int, status):
    if not item_Bytes:
        raise ValueError("item_bytes cannot be empty")
    packet_len = self.write_page_len
    packet_count = math.ceil(len(item_Bytes) / packet_len)
    data_code = 0   # 0-write 2-read
    if packet_count == 0:
        return True
    for i in range(0, packet_count):
        current_packet_bytes = get_send_packet_bytes(
                item, i, data_code, packet_count,
                item_Bytes[i * packet_len: (i + 1) * packet_len])
        flag, new_data_bytes = exchange_block_with_radio(
            current_packet_bytes,
            serial, self.read_packet_len)
        if (not flag
           or (len(new_data_bytes) > 14 and new_data_bytes[14] != item)):
            raise errors.RadioError("radio reported failure")
        status.cur += len(current_packet_bytes)
        self.status_fn(status)


def get_write_item_bytes(all_bytes: bytearray, item):
    """
       Get the bytes for a specific memory region
       - Uses MEMORY_REGIONS_RANGES to find start and length
    """
    if item not in MEMORY_REGIONS_RANGES:
        LOG.debug(f"Unknown memory Range:{item}")
        return b""
    start, length = MEMORY_REGIONS_RANGES[item]
    return all_bytes[start:start+length]


def write_memory_region(all_bytes: bytearray,
                        item_bytes: bytes, item):
    """
    Put the given region bytes into the main memory array
    - Uses MEMORY_REGIONS_RANGES to find start and length
    """
    if item not in MEMORY_REGIONS_RANGES:
        LOG.debug(f"Unknown memory Range:{item}")
        return
    start, length = MEMORY_REGIONS_RANGES[item]
    item_len = min(len(item_bytes), length)
    all_bytes[start:start + item_len] = item_bytes[:item_len]


def exit_programming_mode(self):
    serial = self.pipe
    """
    exiting read/write mode and reboot radio
    """
    reboot_sign = 111
    packet_index = 0
    data_code = 0   # 0-write 2-read
    packet_count = 1
    data_buffer = b'\x00'
    try:
        serial.write(get_send_packet_bytes(reboot_sign,
                                           packet_index, data_code,
                                           packet_count, data_buffer))
    except Exception as e:
        LOG.debug(
            f"Radio refused to exit programming mode:{e}")
        raise errors.RadioError(
            "Radio refused to exit programming mode")


def get_send_packet_bytes(data_type: int, packet_index: int,
                          data_code: int, packet_count: int,
                          data_buffer: bytes):
    data_len = len(data_buffer)
    data_part = struct.pack(f'<12sHBBHH{data_len}s',
                            b"RDTP\x01\x00\x00\x00\x00\x00\x00\x00",
                            data_len + 6,
                            data_type & 0xFF,
                            data_code & 0xFF,  # 0-write 2-read
                            packet_index & 0xFFFF,
                            packet_count & 0xFFFF,
                            data_buffer)
    return data_part + calculate_crc16(data_part) + b"\xff"


def calculate_crc16(buf):
    crc = 0x0000
    for b in buf:
        crc ^= b
        for _ in range(8):
            if crc & 0x0001:
                crc = (crc >> 1) ^ 0xA001
            else:
                crc >>= 1
    return struct.pack("<H", crc)


def _get_memory(self, mem, _mem, ch_index):
    mem.extra = RadioSettingGroup("Extra", "extra")
    mem.extra.append(
        RadioSetting(
            "tottime",
            "TOT",
            RadioSettingValueList(
                get_namedict_by_items(TIMEOUTTIMER_LIST),
                current_index=(_mem.tottime - 1))))
    mem.extra.append(
        RadioSetting(
            "totpermissions",
            "TX Permissions",
            RadioSettingValueList(TOTPERMISSIONS_LIST,
                                  current_index=_mem.totpermissions)))
    mem.extra.append(
        RadioSetting(
            "rxsqlmode",
            "Squelch Level",
            RadioSettingValueList(SQUELCHLEVEL_LIST,
                                  current_index=_mem.rxsqlmode)))
    mem.extra.append(
        RadioSetting(
            "autoscan",
            "Auto Scan System",
            RadioSettingValueList(AUTOSCAN_LIST,
                                  current_index=_mem.autoscan)))
    mem.extra.append(
        RadioSetting(
            "offlineorreversal",
            "Talkaround & Reversal",
            RadioSettingValueList(OFFLINE_REVERSAL_LIST,
                                  current_index=_mem.offlineorreversal)))
    mem.extra.append(
        RadioSetting(
            "vox", "VOX",
            RadioSettingValueBoolean(_mem.vox)))
    mem.extra.append(
        RadioSetting(
            "companding", "Compander",
            RadioSettingValueBoolean(_mem.companding)))
    mem.extra.append(
        RadioSetting("scramble", "Scramble",
                     RadioSettingValueBoolean(_mem.scramble)))
    ch_index_dict = get_ch_index(self)
    if ch_index not in ch_index_dict:
        mem.freq = 0
        mem.empty = True
        return mem
    mem.freq = _mem.rxfreq
    mem.name = self.filter_name(str(_mem.alias).rstrip())
    tx_freq = _mem.txfreq
    if mem.freq == 0:
        mem.empty = True
        return mem
    if mem.freq == 0xFFFFFFFF:
        mem.freq = 0
        mem.empty = True
        return mem
    if mem.freq == tx_freq:
        mem.duplex = ""
        mem.offset = 0
    elif tx_freq == 0xFFFFFFFF:
        mem.duplex = ""
        mem.offset = 0
    else:
        mem.duplex = mem.freq > tx_freq and "-" or "+"
        mem.offset = abs(mem.freq - tx_freq)
    mem.mode = BANDWIDEH_LIST[(1 if _mem.bandwidth >= 3 else 0)]
    rxtone = txtone = None
    if _mem.rxctcvaluetype == 1:
        tone_value = _mem.rxctc / 10.0
        if tone_value in chirp_common.TONES:
            rxtone = tone_value
    elif _mem.rxctcvaluetype in [2, 3]:
        rxtone = int("%03o" % _mem.rxctc)
    if _mem.txctcvaluetype == 1:
        tone_value = _mem.txctc / 10.0
        if tone_value in chirp_common.TONES:
            txtone = tone_value
    elif _mem.txctcvaluetype in [2, 3]:
        txtone = int("%03o" % _mem.txctc)
    rx_tone = (("" if _mem.rxctcvaluetype == 0
                else "Tone" if _mem.rxctcvaluetype == 1 else "DTCS"),
               rxtone, (_mem.rxctcvaluetype == 0x3) and "R" or "N")
    tx_tone = (("" if _mem.txctcvaluetype == 0
                else "Tone" if _mem.txctcvaluetype == 1 else "DTCS"),
               txtone, (_mem.txctcvaluetype == 0x3) and "R" or "N")
    chirp_common.split_tone_decode(mem, tx_tone, rx_tone)
    mem.power = POWER_LEVELS[(1 if _mem.power == 2 else 0)]
    if ch_index < 2:
        mem.immutable += ["name"]
    return mem


def get_model_info(self, model_info):
    rs_value = RadioSettingValueString(0, 20, self.current_model)
    rs_value.set_mutable(False)
    rs = RadioSetting("modelinfo.Machinecode", "Machine Code", rs_value)
    model_info.append(rs)
    rs_value = RadioSettingValueString(
        0, 100,
        "136.00000-174.00000, 400.00000-480.00000")
    rs_value.set_mutable(False)
    rs = RadioSetting("modelinfo.freqrange", "Frequency Range[MHz]", rs_value)
    model_info.append(rs)


def get_common_setting(self, common):
    _settings = self._memobj.settings
    opts = ["Low", "Normal", "Strengthen"]
    common.append(RadioSetting(
        "settings.micmain", "Mic Main",
        RadioSettingValueList(opts, current_index=_settings.micmain)))
    opts = ["Stun WakeUp", "Stun TX", "Stun TX/RX"]
    common.append(
        RadioSetting("settings.stunmode", "Stun Type",
                     RadioSettingValueList(opts,
                                           current_index=_settings.stunmode)))
    opts_dict = [{"name": "%s" % (x + 1), "id": x} for x in range(0, 10, 1)]
    common.append(get_radiosetting_by_key(self, _settings,
                                          "calltone", "Call Tone",
                                          _settings.calltone, opts_dict))
    common.append(
        get_radiosetting_by_key(
            self, _settings, "freqstep", "Frequency Step",
            _settings.freqstep, FREQ_STEP_List))
    opts = ["OFF", "1:1", "1:2", "1:4"]
    common.append(
        RadioSetting(
            "settings.powersavingmode", "Battery Mode",
            RadioSettingValueList(
                opts, current_index=_settings.powersavingmode)))
    opts_dict = [
        {"name": "%ss" % ((x + 1) * 5), "id": x} for x in range(0, 16, 1)]
    common.append(
        get_radiosetting_by_key(
            self, _settings, "powersavingdelaytime", "Battery Delay Time",
            _settings.powersavingdelaytime, opts_dict))
    opts_dict = [{"name": "%s" % x, "id": x} for x in range(1, 16, 1)]
    common.append(
        get_radiosetting_by_key(
            self, _settings, "backlightbrightness", "Backlight Brightness",
            _settings.backlightbrightness, opts_dict))

    opts = [
        "Always", "5s", "10s", "15s", "20s", "25s", "30s",
        "1min", "2min", "3min", "4min", "5min", "15min", "30min",
        "45min", "1h"]
    common.append(
        RadioSetting("settings.backlighttime", "Backlight Time",
                     RadioSettingValueList(
                         opts, current_index=_settings.backlighttime)))
    opts_dict = [{"name": "%s" % x, "id": x} for x in range(1, 16, 1)]
    common.append(
        get_radiosetting_by_key(
            self, _settings, "voxthreshold", "VOX Level",
            _settings.voxthreshold, opts_dict))
    opts_dict = [
        {"name": "%sms" % ((x) * 500), "id": x} for x in range(1, 5, 1)]
    common.append(
        get_radiosetting_by_key(
            self, _settings, "voxdelaytime", "VOX Delay Time",
            _settings.voxdelaytime, opts_dict))
    opts_dict = [{"name": "OFF", "id": 0}] + [
        {"name": "%ss" % x, "id": x} for x in range(5, 256, 5)]

    common.append(
        get_radiosetting_by_key(
            self, _settings, "menuouttime", "Menu Timeout Setting",
            _settings.menuouttime, opts_dict))
    opts = ["Manual", "Auto"]
    common.append(
        RadioSetting(
            "settings.autoormanuallock", "Key Lock Mode",
            RadioSettingValueList(opts,
                                  current_index=_settings.autoormanuallock)))
    opts = ["Frequency", "Name", "Channel"]
    common.append(
        RadioSetting(
            "settings.chdisplay", "Display Mode",
            RadioSettingValueList(opts, current_index=_settings.chdisplay)))
    opts = ["Band A", "Band B", "Band A & Band B", "Band B & Band A"]
    rs = RadioSetting(
        "settings.homeselect",
        "Band Selection",
        RadioSettingValueList(
            opts,
            current_index=get_band_selection(
                _settings.homeselect, _settings.homeindex)))
    rs.set_apply_callback(
        set_band_selection, _settings, opts, "homeselect", "homeindex")
    common.append(rs)
    opts = ["Channel", "VFO Frequency"]
    common.append(
        RadioSetting(
            "settings.homechtype_1",
            "Channel Type A",
            RadioSettingValueList(opts, current_index=_settings.homechtype_1)))
    common.append(
        RadioSetting(
            "settings.homechtype_2",
            "Channel Type B",
            RadioSettingValueList(opts, current_index=_settings.homechtype_2)))
    opts = ["Last Active Channel", "Designated Channel"]
    common.append(
        RadioSetting(
            "settings.poweron_type_1",
            "Power On A",
            RadioSettingValueList(opts,
                                  current_index=_settings.poweron_type_1)))
    common.append(
        RadioSetting(
            "settings.poweron_type_2",
            "Power On B",
            RadioSettingValueList(opts,
                                  current_index=_settings.poweron_type_2)))
    opts_dict = get_scan_item_list(self)
    common.append(
        get_radiosetting_by_key(
            self, _settings, "scanlist", "Enable Scan List",
            _settings.scanlist, opts_dict))
    short_dict = SIDE_KEY_LIST[:12]
    common.append(
        get_radiosetting_by_key(
            self, _settings, "tk1_short", "Top Key Short Press",
            _settings.tk1_short, short_dict))
    common.append(
        get_radiosetting_by_key(
            self, _settings, "tk1_long", "Top Key Long Press",
            _settings.tk1_long, short_dict))
    common.append(
        get_radiosetting_by_key(
            self, _settings, "sk1_short", "Short Press Side Key 1",
            _settings.sk1_short, short_dict))
    common.append(
        get_radiosetting_by_key(
            self, _settings, "sk1_long", "Long Press Side Key 1",
            _settings.sk1_long, SIDE_KEY_LIST))
    common.append(
        get_radiosetting_by_key(
            self, _settings, "sk2_short", "Short Press Side Key 2",
            _settings.sk2_short, short_dict))
    common.append(
        get_radiosetting_by_key(
            self, _settings, "sk2_long", "Long Press Side Key 2",
            _settings.sk2_long, SIDE_KEY_LIST))
    opts = ["55Hz", "120°", "180°", "240°"]
    common.append(
        RadioSetting(
            "settings.tailsoundeliminationsfre", "CTC Tail Elimination",
            RadioSettingValueList(
                opts, current_index=_settings.tailsoundeliminationsfre)))
    if _settings.salezone != 2:
        opts = ["NOAA-%s" % x for x in range(1, 13, 1)]
        common.append(
            RadioSetting(
                "settings.wxch", "NOAA Channel",
                RadioSettingValueList(opts, current_index=_settings.wxch)))
    opts = ["1000Hz", "1450Hz", "1750Hz", "2100Hz"]
    common.append(
        RadioSetting(
            "settings.singletone", "Single Tone",
            RadioSettingValueList(opts, current_index=_settings.singletone)))
    common.append(
        RadioSetting(
            "settings.tailsoundeliminationswitch",
            "Tail Elimination Switch",
            RadioSettingValueBoolean(_settings.tailsoundeliminationswitch)))
    common.append(
        RadioSetting(
            "settings.rogerbeep", "Roger Beep",
            RadioSettingValueBoolean(_settings.rogerbeep)))
    common.append(
        RadioSetting(
            "settings.touchtone", "Key Beep",
            RadioSettingValueBoolean(_settings.touchtone)))
    common.append(
        RadioSetting(
            "settings.txpermittone", "TX Permit Tone",
            RadioSettingValueBoolean(_settings.txpermittone)))
    common.append(
        RadioSetting(
            "settings.voiceprompts", "Voice Broadcast",
            RadioSettingValueBoolean(_settings.voiceprompts)))
    common.append(
        RadioSetting(
            "settings.chknoblock", "Channel Knob Lock",
            RadioSettingValueBoolean(_settings.chknoblock)))
    common.append(
        RadioSetting(
            "settings.panellock", "Keyboard Lock",
            RadioSettingValueBoolean(_settings.panellock)))

    common.append(
        RadioSetting(
            "settings.sidekeylock", "Side Key Lock",
            RadioSettingValueBoolean(_settings.sidekeylock)))


def get_dtmf_setting(self, dtmf):
    _dtmf_comm = self._memobj.dtmfcomm
    opts = ["OFF"] + ["%ss" % x for x in range(1, 16, 1)]
    dtmf.append(
        RadioSetting(
            "dtmfsetting.autoresettime",
            "Auto Reset Time",
            RadioSettingValueList(opts,
                                  current_index=_dtmf_comm.autoresettime)))
    opts = [{"name": "Stun TX", "id": 1}, {"name": "Stun TX/RX", "id": 2}]
    dtmf.append(
        get_radiosetting_by_key(
            self, _dtmf_comm, "stunmode",
            "Stun Type", _dtmf_comm.stunmode, opts))
    opts = ["300ms", "550ms", "800ms", "1050ms"]
    dtmf.append(
        RadioSetting(
            "dtmfsetting.codedelaytime", "Digit Delay",
            RadioSettingValueList(opts,
                                  current_index=_dtmf_comm.codedelaytime)))
    opts = ["OFF", "BOT", "EOT", "Both"]
    dtmf.append(
        RadioSetting(
            "dtmfsetting.pttidtype", "PTT ID Type",
            RadioSettingValueList(opts,
                                  current_index=_dtmf_comm.pttidtype)))
    dtmf.append(
        RadioSetting(
            "dtmfsetting.showani", "Show ANI",
            RadioSettingValueBoolean(_dtmf_comm.showani)))
    dtmf.append(
        RadioSetting(
            "dtmfsetting.sidetone", "Side Tone",
            RadioSettingValueBoolean(_dtmf_comm.sidetone)))
    dtmf.append(
        RadioSetting(
            "dtmfsetting.callid", "Call ID",
            RadioSettingValueString(
                0, 10,
                "".join(filter(_dtmf_comm.callid, DTMFCHARSET, 10, True)),
                False, DTMFCHARSET)))
    dtmf.append(
        RadioSetting(
            "dtmfsetting.stunid", "Stun ID",
            RadioSettingValueString(
                0, 10,
                "".join(filter(_dtmf_comm.stunid, DTMFCHARSET, 10, True)),
                False, DTMFCHARSET)))
    dtmf.append(
        RadioSetting(
            "dtmfsetting.revive", "Revive ID",
            RadioSettingValueString(
                0, 10,
                "".join(filter(_dtmf_comm.revive, DTMFCHARSET, 10, True)),
                False, DTMFCHARSET)))
    dtmf.append(
        RadioSetting(
            "dtmfsetting.bot", "BOT",
            RadioSettingValueString(
                0, 16,
                "".join(filter(_dtmf_comm.bot, DTMFCHARSET, 16, True)),
                False, DTMFCHARSET)))
    dtmf.append(
        RadioSetting(
            "dtmfsetting.eot", "EOT",
            RadioSettingValueString(
                0, 16,
                "".join(filter(_dtmf_comm.eot, DTMFCHARSET, 16, True)),
                False, DTMFCHARSET)))


def get_vfo_scan(self, vfoscan):
    _vfo_scan = self._memobj.vfoscans[0]
    opts = ["Carrier", "Time", "Search"]
    vfoscan.append(
        RadioSetting(
            "vfoscan.scanmode", "Scan Mode",
            RadioSettingValueList(opts, current_index=_vfo_scan.scanmode)))
    opts = ["Carrier", "CTC/DCS"]
    vfoscan.append(
        RadioSetting(
            "vfoscan.scancondition", "Scan Condition",
            RadioSettingValueList(opts,
                                  current_index=_vfo_scan.scancondition)))

    opts = ["%s" % (x + 1) for x in range(0, 16, 1)]
    vfoscan.append(
        RadioSetting(
            "vfoscan.hangtime", "Scan Hang Time[s]",
            RadioSettingValueList(opts, current_index=_vfo_scan.hangtime)))
    vfoscan.append(
        RadioSetting(
            "vfoscan.talkback", "Talk Back Enable",
            RadioSettingValueBoolean(_vfo_scan.talkback)))
    opts = ["Current Frequency", "Starting Frequency"]
    vfoscan.append(
        RadioSetting(
            "vfoscan.startcondition", "Start Condition",
            RadioSettingValueList(opts,
                                  current_index=_vfo_scan.startcondition)))

    freq_start = _vfo_scan.vhffreq_start / 1000000
    vfoscan.append(
        RadioSetting(
            "vfoscan.vhffreq_start", "Start Frequency",
            RadioSettingValueFloat(136, 480, freq_start, 0.00001, 5)))

    freq_end = _vfo_scan.vhffreq_end / 1000000
    vfoscan.append(
        RadioSetting(
            "vfoscan.vhffreq_end", "End Frequency",
            RadioSettingValueFloat(136, 480, freq_end, 0.00001, 5)))


def _set_memory(self, mem, _mem, ch_index):
    ch_index_dict = get_ch_index(self)
    rx_freq = get_ch_rxfreq(mem)
    flag = ch_index not in ch_index_dict and rx_freq != 0
    if rx_freq != 0xFFFFFFFF and flag:
        ch_index_dict.append(ch_index)
    elif ch_index in ch_index_dict and rx_freq <= 0:
        ch_index_dict.remove(ch_index)
    set_ch_index(self, ch_index_dict)
    _mem.set_raw(b"\x00" * 40)
    if mem.empty:
        return
    _mem.rxfreq = rx_freq
    _mem.alias = mem.name.ljust(14)
    txfrq = (int(rx_freq - mem.offset)
             if mem.duplex == "-" and mem.offset > 0
             else (int(rx_freq + mem.offset)
                   if mem.duplex == "+" and mem.offset > 0
                   else rx_freq))
    _mem.txfreq = txfrq
    _mem.bandwidth = 3 if mem.mode == "FM" else 1
    if mem.power in POWER_LEVELS:
        _mem.power = 2 if POWER_LEVELS.index(mem.power) == 1 else 0
    else:
        _mem.power = 0
    ((txmode, txtone, txpol),
     (rxmode, rxtone, rxpol)) = chirp_common.split_tone_encode(
        mem)
    if rxmode == "Tone":
        _mem.rxctcvaluetype = 1
        rxtone_value = int(rxtone * 10)
        _mem.rxctc = rxtone_value
    elif rxmode == "DTCS":
        _mem.rxctcvaluetype = rxpol == "R" and 3 or 2
        oct_value = int(str(rxtone), 8)
        _mem.rxctc = oct_value
    else:
        _mem.rxctcvaluetype = 0
    if txmode == "Tone":
        _mem.txctcvaluetype = 1
        txtone_value = int(txtone * 10)
        _mem.txctc = txtone_value
    elif txmode == "DTCS":
        _mem.txctcvaluetype = txpol == "R" and 3 or 2
        oct_value = int(str(txtone), 8)
        _mem.txctc = oct_value
    else:
        _mem.txctcvaluetype = 0
    for setting in mem.extra:
        if setting.get_name() == "tottime":
            _mem.tottime = get_item_by_name(
                TIMEOUTTIMER_LIST, str(setting.value))
        else:
            setattr(_mem, setting.get_name(), setting.value)
    return _mem


def get_radiosetting_by_key(self, _setting, item_key,
                            item_display_name, item_value,
                            opts_dict, callback=None):
    opts = get_namedict_by_items(opts_dict)
    item = RadioSetting(
        item_key, item_display_name,
        RadioSettingValueList(
            opts, current_index=get_item_by_id(opts_dict, item_value)))
    item.set_apply_callback(
        set_item_callback if callback is None else callback,
        _setting, item_key, opts_dict, self)
    return item


def get_namedict_by_items(items):
    return [x["name"] for x in items]


def get_item_by_id(items, value):
    return next((i for i, item in enumerate(items) if item["id"] == value), 0)


def set_item_callback(set_item, obj, name, items, self):
    item_value = get_item_by_name(items, set_item.value)
    setattr(obj, name, item_value)


def get_item_by_name(items, name):
    return next(
        (item["id"] for item in items if item["name"] == name),
        items[0]["id"] if items else 0)


def get_band_selection(band_value, band_index):
    return (
        (3 if band_value >= 3 else 1)
        if band_index >= 1
        else (2 if band_value >= 3 else 0))


def get_ch_index(self):
    if self._ch_cache is None:
        ch_data = self._memobj.channeldata
        self._ch_cache = []
        seen = set()
        for i in range(ch_data.chnum):
            ch_index = int(ch_data.chindex[i])
            if ch_index != 0xFFFF and ch_index not in seen:
                seen.add(ch_index)
                self._ch_cache.append(ch_index)
    return self._ch_cache


def set_ch_index(self, ch_index_list):
    if not isinstance(ch_index_list, list):
        raise TypeError("ch_index must be a list")
    ch_index_list = sorted(int(x) for x in ch_index_list)
    _ch_data = self._memobj.channeldata
    ch_num = len(ch_index_list)
    if len(_ch_data.chindex) < ch_num:
        raise ValueError("Not enough space in chindex array")
    _ch_data.chnum = ch_num
    for i in range(1027):
        _ch_data.chindex[i] = (
            ch_index_list[i] if i < ch_num else 0xFFFF)
    self._ch_cache = ch_index_list.copy()


def get_ch_rxfreq(mem):
    if mem.empty:
        return 0
    ch_freq = (mem.freq // 10) * 10
    if mem.freq == 0:
        return mem.freq
    elif mem.freq < 136000000:
        ch_freq = 136000000
    elif mem.freq > 174000000 and mem.freq < 400000000:
        ch_freq = 174000000
    elif mem.freq > 480000000:
        ch_freq = 480000000
    return ch_freq


def get_scan_item_list(self):
    _scandata = self._memobj.scandata
    _scans = self._memobj.scans
    scan_dict = []
    scan_num = _scandata.scannum
    if scan_num > 0:
        for i in range(0, scan_num):
            scan_index = _scandata.scanindex[i]
            if scan_index < 16:
                scan_item = _scans[scan_index]
                scanname = "".join(filter(scan_item.name, NAMECHARSET, 12))
                scan_dict.append({"name": scanname, "id": scan_index})
    return scan_dict


def set_band_selection(set_item, obj,
                       opts, home_select_field_name,
                       home_index_field_name):
    value = opts.index(set_item.value)
    value_mapping = {3: (1, 3), 2: (0, 3), 1: (1, 2), 0: (0, 1)}
    home_index, home_select = value_mapping.get(value, (0, 1))
    setattr(obj, home_index_field_name, home_index)
    setattr(obj, home_select_field_name, home_select)


def filter(s, char_set, max_length=10, is_upper=False):
    s_ = ""
    input_len = len(s)
    for i in range(0, min(max_length, input_len)):
        c = str(s[i])
        if is_upper:
            c = c.upper()
        s_ += c if c in char_set else ""
    return s_


@directory.register
class HA1G(chirp_common.CloneModeRadio):
    """Retevis HA1G"""

    VENDOR = "Retevis"
    MODEL = "HA1G"
    BAUD_RATE = 115200
    _memsize = 0xD868
    read_packet_len = 1047
    write_page_len = 1024
    current_model = "HA1G"
    _ch_cache = None

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.valid_special_chans = sorted(SPECIAL_MEMORIES.keys())
        rf.memory_bounds = (1, 256)  # Channel range
        rf.has_ctone = True
        rf.has_rx_dtcs = True
        rf.has_dtcs = True
        rf.has_cross = True
        rf.has_bank = False
        rf.valid_bands = [(136000000, 174000010), (400000000, 480000010)]
        rf.has_tuning_step = False
        rf.has_nostep_tuning = True
        rf.valid_name_length = 12
        rf.valid_skips = []
        rf.valid_tuning_steps = []
        rf.valid_characters = chirp_common.CHARSET_UPPER_NUMERIC + "".join(
            c for c in "-/;,._!? *#@$%&+=/<>~(){}]'"
            if c not in chirp_common.CHARSET_UPPER_NUMERIC)
        rf.has_settings = True
        rf.valid_modes = BANDWIDEH_LIST
        rf.valid_power_levels = POWER_LEVELS
        rf.has_comment = True
        rf.valid_tmodes = ["", "Tone", "TSQL", "DTCS", "Cross"]
        rf.valid_cross_modes = [
            "Tone->Tone",
            "Tone->DTCS",
            "DTCS->Tone",
            "->Tone",
            "->DTCS",
            "DTCS->",
            "DTCS->DTCS"]
        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def sync_in(self):
        try:
            self._mmap = do_download(self)
            self.process_mmap()
        except errors.RadioError:
            raise
        except Exception as e:
            raise errors.RadioError(
                "Failed to communicate with radio: %s" % e)

    def sync_out(self):
        """Upload to radio"""
        try:
            logging.debug("come in sync_out")
            do_upload(self)
        except Exception as e:
            LOG.exception(
                "Unexpected error during upload: %s" % e)
            raise errors.RadioError(
                "Unexpected error communicating with the radio")

    def get_memory(self, number):
        mem = chirp_common.Memory()
        ch_index = 0
        if isinstance(number, str):
            mem.extd_number = number
            ch_index = 0 if number == "VFOA" else 1
            mem.number = len(self._memobj.channels) + ch_index + 1
        elif number > len(self._memobj.channels):
            mem.extd_number = (
                number - len(self._memobj.channels) == 1 and "VFOA" or "VFOB")
            number = mem.extd_number
            ch_index = 0 if number == "VFOA" else 1
        else:
            ch_index = number + 2
            mem.number = number
        _mem = self._memobj.channels[ch_index]
        mem = _get_memory(self, mem, _mem, ch_index)
        if ch_index > 2 and ch_index < 33:
            mem.immutable = ["empty", "freq", "duplex", "offset"]
        if ch_index >= 10 and ch_index < 17:
            mem.immutable += ["mode", "power"]
        return mem

    def set_memory(self, mem):
        ch_index = 0
        if mem.number > len(self._memobj.channels):
            ch_index = 0 if mem.extd_number == "VFOA" else 1
        else:
            ch_index = mem.number + 2
        _mem = self._memobj.channels[ch_index]
        if ch_index < 33 and mem.freq == 0:
            return
        _set_memory(self, mem, _mem, ch_index)

    def get_raw_memory(self, number):
        if isinstance(number, str):
            ch_index = 0 if number == "VFOA" else 1
        else:
            ch_index = number + 2
        _mem = self._memobj.channels[ch_index]
        return repr(_mem)

    def get_settings(self):
        ModelInfo = RadioSettingGroup("info", "Model Information")
        common = RadioSettingGroup("basic", "Common Settings")
        dtmf = RadioSettingGroup("dtmfe", "DTMF Settings")
        vfoscan = RadioSettingGroup("vfoscan", "VFO  Scan")
        setmode = RadioSettings(ModelInfo, common, dtmf, vfoscan)
        try:
            get_model_info(self, ModelInfo)
            get_common_setting(self, common)
            get_dtmf_setting(self, dtmf)
            get_vfo_scan(self, vfoscan)
        except Exception as e:
            LOG.exception("Error getting settings: %s", e)
        return setmode

    def set_settings(self, uisettings):
        for element in uisettings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            if not element.changed():
                continue
            try:
                if element.has_apply_callback():
                    LOG.debug("Using apply callback")
                    element.run_apply_callback()
                else:
                    name = element.get_name()
                    value = element.value
                    if name.startswith("settings."):
                        name = name[9:]
                        _settings = self._memobj.settings
                        setattr(_settings, name, value)
                    elif name.startswith("dtmfsetting."):
                        name = name[12:]
                        _dtmfcomm = self._memobj.dtmfcomm
                        if (name in ["callid", "stunid", "revive"]):
                            value = filter(value, DTMFCHARSET, 10, True)
                            value = value.ljust(10, "\x00")
                        elif name in ["bot", "eot"]:
                            value = filter(value, DTMFCHARSET, 16, True)
                            value = value.ljust(16, "\x00")
                        setattr(_dtmfcomm, name, value)
                    elif name.startswith("vfoscan."):
                        name = name[8:]
                        _vfo_scan = self._memobj.vfoscans[0]
                        if name in ["vhffreq_start", "vhffreq_end"]:
                            value = int(value * 1000000)
                        setattr(_vfo_scan, name, value)
                    LOG.debug("Setting %s: %s", name, value)
            except Exception:
                LOG.exception(element.get_name())
                raise


@directory.register
class HA1UV(HA1G):
    """Retevis HA1UV"""

    MODEL = "HA1UV"
    current_model = "HA1UV"

    def get_features(self):
        rf = super().get_features()
        rf.memory_bounds = (1, 1024)  # Channel range
        return rf

    def validate_memory(self, mem):
        return super().validate_memory(mem)

    def get_memory(self, number):
        mem = chirp_common.Memory()
        ch_index = 0
        if isinstance(number, str):
            mem.extd_number = number
            ch_index = 0 if number == "VFOA" else 1
            mem.number = len(self._memobj.channels) + ch_index + 1
        elif number > len(self._memobj.channels):
            mem.extd_number = (
                number - len(self._memobj.channels) == 1 and "VFOA" or "VFOB")
            number = mem.extd_number
            ch_index = 0 if number == "VFOA" else 1
        else:
            ch_index = number + 2
            mem.number = number
        _mem = self._memobj.channels[ch_index]
        return _get_memory(self, mem, _mem, ch_index)

    def set_memory(self, mem):
        ch_index = 0
        if mem.number > len(self._memobj.channels):
            ch_index = 0 if mem.extd_number == "VFOA" else 1
        else:
            ch_index = mem.number + 2
        _mem = self._memobj.channels[ch_index]
        if ch_index < 2 and mem.freq == 0:
            return
        _set_memory(self, mem, _mem, ch_index)
        LOG.debug("Setting %i(%s)" % (mem.number, mem.extd_number))



# Copyright 2024 Marcos Del Sol Vives <marcos@orca.pet>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import namedtuple
import logging
import struct
import time
from chirp import chirp_common, directory, errors, memmap
from chirp import bitwise
from chirp.drivers import baofeng_common
from chirp.settings import (RadioSetting, RadioSettingGroup,
                            RadioSettingValueBoolean,
                            RadioSettingValueList)


LOG = logging.getLogger(__name__)

CMD_ACK = b'\x06'
CMD_READ = b'R'
CMD_WRITE = b'W'
CMD_EXIT = b'E'

MemoryChunk = namedtuple('MemoryChunk', ['start', 'end', 'read_only'])

ProtoVersion = namedtuple('ProtoVersion',
                          ['chan_count', 'serial_magic', 'mem_chunks'])

PROTO_A = ProtoVersion(
    chan_count=16,
    serial_magic=b'\x80\x82',
    mem_chunks=[
        MemoryChunk(start=0x000, end=0x10F, read_only=False),
        MemoryChunk(start=0x2B0, end=0x2BF, read_only=False),
        MemoryChunk(start=0x330, end=0x33F, read_only=True),
        MemoryChunk(start=0x380, end=0x3DF, read_only=False),
    ],
)
PROTO_B = ProtoVersion(
    chan_count=16,
    serial_magic=b'\x80\x82',
    mem_chunks=PROTO_A.mem_chunks,
)
PROTO_C = ProtoVersion(
    chan_count=16,
    serial_magic=b'\x20\x22',
    mem_chunks=PROTO_A.mem_chunks,
)
PROTO_D = ProtoVersion(
    chan_count=30,
    serial_magic=b'\x80\x82',
    mem_chunks=[
        MemoryChunk(start=0x000, end=0x1EF, read_only=False),
        MemoryChunk(start=0x2B0, end=0x2BF, read_only=False),
        MemoryChunk(start=0x330, end=0x33F, read_only=True),
        MemoryChunk(start=0x380, end=0x3DF, read_only=False),
    ],
)


class BaofengDigital(chirp_common.CloneModeRadio,
                     chirp_common.ExperimentalRadio):
    """Base for all digital radios from Fujian Baofeng Electronics Co. Ltd."""
    VENDOR = "Baofeng"
    BAUD_RATE = 9600

    MEM_FORMAT = """
        #seekto 0x0010;
        struct {{
            lbcd rx_freq[4];
            lbcd tx_freq[4];
            lbcd rx_tone[2];
            lbcd tx_tone[2];
            u8 hopping:1,
               unused1:2,
               scan:1,
               high_power:1,
               narrow:1,
               digital:1,
               bcl_disabled:1;
            u8 unused2:3,
               encryption_key:5;
            u16 unused3;
        }} memory[{chan_count}];
    """

    _PREAMBLE_CMDS = [
        (b'\x02', b'P3107\xf7\x00\x00'),
        (CMD_ACK, CMD_ACK),
        (b'R\x01\x30\x08', b'W\x01\x30\x08\xff\xff\xff\xff\xff\xff\xff\xff'),
        (CMD_ACK, CMD_ACK),
    ]

    _WRITE_PREAMBLE_CMDS = [
        (b'E', b'F'),
        (b'\x02\x80\x82OGRAM', CMD_ACK),
        (b'\x02', b'P3107\xf7\x00\x00'),
        (CMD_ACK, CMD_ACK),
    ]

    _VOICES_PER_REGION = {
        0x00: ('English', 'Chinese'),
        0x01: ('Chinese', 'Russian'),
        0x02: ('Russian', 'English'),
        0x03: ('Arabic', 'English'),
        0x04: ('Arabic', 'Chinese'),
        0x05: ('Arabic', 'Russian'),
        0x06: ('Chinese', 'Portuguese'),
        0x07: ('English', 'Portuguese'),
        0x08: ('Chinese', 'Spanish'),
        0x09: ('English', 'Spanish'),
        0x0A: ('Chinese', 'German'),
        0x0B: ('English', 'German'),
        0x0C: ('Chinese', 'French'),
        0x0D: ('English', 'French'),
    }

    _ENC_OPTS = ['None (analog)'] + [str(x) for x in range(32)]

    @property
    def _memsize(self):
        return self._proto.mem_chunks[-1].end + 1

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_rx_dtcs = True
        rf.has_bank = False
        rf.has_name = False
        rf.has_tuning_step = False
        rf.can_odd_split = True
        rf.has_cross = True
        rf.memory_bounds = (1, self._proto.chan_count)
        rf.valid_bands = self.VALID_BANDS
        rf.valid_duplexes = ["", "+", "-", "split"]
        rf.valid_modes = ['FM', 'NFM']
        rf.valid_power_levels = self.POWER_LEVELS
        rf.valid_tuning_steps = [2.5, 6.25, 5.0]
        rf.valid_tmodes = ["", "Tone", "TSQL", "DTCS", "Cross"]
        rf.valid_cross_modes = [
            "Tone->Tone",
            "Tone->DTCS",
            "DTCS->Tone",
            "DTCS->",
            "->Tone",
            "->DTCS",
            "DTCS->DTCS",
        ]
        return rf

    def _get_radio_id(self):
        for attempt in range(5):
            try:
                baofeng_common._clean_buffer(self)

                self.pipe.write(b'\x02prOGRAM')
                resp = self.pipe.read(4)
                if len(resp) != 4:
                    raise errors.RadioError(
                        'Radio did not send identification')
                return resp
            except errors.RadioError:
                # Don't wait if already last attempt
                if attempt == 4:
                    raise
                self._exit_programming()
                time.sleep(1)

    def _enter_programming(self, write):
        radio_id = self._get_radio_id()
        if radio_id[1] != self._serial_id:
            raise errors.RadioError('Detected a different radio model')

        self._metadata['fw_region'] = radio_id[0]
        self._metadata['fw_version'] = radio_id[2] * 0x100 + radio_id[3]

        # Issue protocol-specific command
        self.pipe.write(b'\x02' + self._proto.serial_magic + b'OGRAM')
        if self.pipe.read(1) != CMD_ACK:
            raise errors.RadioError('No ACK on protocol-specific command')

        # Issue fixed commands and check their values
        if write:
            preamble = self._PREAMBLE_CMDS + self._WRITE_PREAMBLE_CMDS
        else:
            preamble = self._PREAMBLE_CMDS
        for step, cmd in enumerate(preamble, 1):
            self.pipe.write(cmd[0])
            if self.pipe.read(len(cmd[1])) != cmd[1]:
                raise errors.RadioError(
                    f'Radio replied incorrectly to preamble step {step}')

    def _clean_buffer(self):
        self.pipe.timeout = 0.001
        self.pipe.read(256)
        self.pipe.timeout = 0.1

    def _exit_programming(self):
        try:
            self.pipe.write(CMD_EXIT)
        except Exception:
            pass

    def sync_in(self):
        try:
            self._enter_programming(False)

            memory = memmap.MemoryMapBytes(b"\xFF" * self._memsize)
            for chunk in self._proto.mem_chunks:
                for addr in range(chunk.start, chunk.end, 8):
                    self.pipe.write(struct.pack('>cHB', CMD_READ, addr, 8))
                    reply = self.pipe.read(12)
                    if len(reply) != 12:
                        raise errors.RadioError(
                            f'Failed to read block at 0x{addr:03x}')

                    self.pipe.write(CMD_ACK)
                    if self.pipe.read(1) != CMD_ACK:
                        raise errors.RadioError(
                            f'No ACK after reading from 0x{addr:03x}')

                    memory.set(addr, reply[4:])
        except errors.RadioError:
            raise
        except Exception as e:
            raise errors.RadioError(
                f'Failed to communicate with the radio: {e}')
        finally:
            self._exit_programming()

        self._mmap = memory
        self.process_mmap()

    def sync_out(self):
        try:
            self._enter_programming(True)

            for chunk in self._proto.mem_chunks:
                if chunk.read_only:
                    continue

                for addr in range(chunk.start, chunk.end, 8):
                    frame = (
                        struct.pack('>cHB', CMD_WRITE, addr, 8) +
                        self._mmap.get(addr, 8)
                    )
                    self.pipe.write(frame)

                    if self.pipe.read(1) != CMD_ACK:
                        raise errors.RadioError(
                            f'Not ACK after writing to 0x{addr:03x}')
        except errors.RadioError:
            raise
        except Exception as e:
            raise errors.RadioError(
                f'Failed to communicate with the radio: {e}')
        finally:
            self._exit_programming()

    def process_mmap(self):
        mem_fmt = self.MEM_FORMAT.format(chan_count=self._proto.chan_count)
        self._memobj = bitwise.parse(mem_fmt, self._mmap)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number])

    def _get_squelch(self, elem):
        # Get raw and check if unset
        if elem.get_raw() in (b'\x00\x00', b'\xFF\xFF'):
            return ('', None, 'N')

        # Get flags, unset them so we can parse as BCD, then restore them
        flags = elem[1].get_bits(0xC0)
        elem[1].clr_bits(0xC0)
        value = int(elem)
        elem[1].set_bits(flags)

        # DTCS if bit 15 is set
        if flags & 0x80:
            return ('DTCS', value, flags & 0x40 and 'R' or 'N')
        else:
            # Parse as CTCSS frequency * 10
            return ('Tone', value / 10.0, 'N')

    def get_memory(self, number):
        raw = self._memobj.memory[number - 1]

        mem = chirp_common.Memory()
        mem.number = number
        mem.empty = raw.rx_freq.get_raw() in (
            b'\x00\x00\x00\x00', b'\xFF\xFF\xFF\xFF'
        )

        bcl = hopping = False
        enc_idx = 0
        if not mem.empty:
            rx_freq = int(raw.rx_freq) * 10
            tx_freq = int(raw.tx_freq) * 10

            mem.freq = rx_freq
            offset = tx_freq - rx_freq
            if offset == 0:
                mem.duplex = ''
                mem.offset = 0
            elif abs(offset) < 10000000:
                mem.duplex = '-' if offset < 0 else '+'
                mem.offset = abs(offset)
            else:
                mem.duplex = 'split'
                mem.offset = tx_freq

            mem.mode = raw.narrow and 'NFM' or 'FM'
            if raw.digital:
                mem.tmode = ''
                mem.dtcs_polarity = 'NN'
                enc_idx = raw.encryption_key + 1
            else:
                hopping = raw.hopping

                tx_tmode, tx_tvalue, tx_tpol = self._get_squelch(raw.tx_tone)
                if tx_tmode == 'Tone':
                    mem.rtone = tx_tvalue
                elif tx_tmode == 'DTCS':
                    mem.dtcs = tx_tvalue

                rx_tmode, rx_tvalue, rx_tpol = self._get_squelch(raw.rx_tone)
                if rx_tmode == 'Tone':
                    mem.ctone = rx_tvalue
                elif rx_tmode == 'DTCS':
                    mem.rx_dtcs = rx_tvalue

                if tx_tmode == 'Tone' and rx_tmode == '':
                    mem.tmode = 'Tone'
                elif tx_tmode and tx_tmode == rx_tmode and \
                        tx_tvalue == rx_tvalue:
                    mem.tmode = tx_tmode == 'Tone' and 'TSQL' or 'DTCS'
                elif tx_tmode or rx_tmode:
                    mem.tmode = 'Cross'
                    mem.cross_mode = f'{tx_tmode}->{rx_tmode}'

                mem.dtcs_polarity = tx_tpol + rx_tpol

            mem.power = self.POWER_LEVELS[raw.high_power]
            mem.skip = '' if raw.scan else 'S'
            bcl = not raw.bcl_disabled

        mem.extra = RadioSettingGroup("Extra", "extra")

        rs = RadioSetting("bcl", "BCL",
                          RadioSettingValueBoolean(bcl))
        rs.set_doc('Busy channel lockout.')
        mem.extra.append(rs)

        rs = RadioSetting("encryption_key", "Encryption key",
                          RadioSettingValueList(self._ENC_OPTS,
                                                current_index=enc_idx))
        rs.set_doc('Encryption key for the digital mode. ' +
                   'If none is selected, analog audio will be used.')
        mem.extra.append(rs)

        rs = RadioSetting("hopping", "Hopping",
                          RadioSettingValueBoolean(hopping))
        rs.set_doc('Frequency hopping. Only supported on analog modes.')
        mem.extra.append(rs)

        return mem

    def _set_no_tone(self, elem):
        elem.set_raw(b'\xFF\xFF')

    def _set_ctcss(self, freq, elem):
        elem.set_value(int(freq * 10))

    def _set_dtcs(self, code, polarity, elem):
        elem.set_value(code)
        elem[1].set_bits(0x80)
        if polarity == 'R':
            elem[1].set_bits(0x40)

    def _set_squelch(self, mode, ctcss_freq, dtcs_code, dtcs_polarity, elem):
        if mode == 'Tone':
            self._set_ctcss(ctcss_freq, elem)
        elif mode == 'DTCS':
            self._set_dtcs(dtcs_code, dtcs_polarity, elem)
        else:
            self._set_no_tone(elem)

    def set_memory(self, mem):
        raw = self._memobj.memory[mem.number - 1]
        if mem.empty:
            raw.fill_raw(b'\xFF')
            return

        raw.fill_raw(b'\x00')

        raw.rx_freq = mem.freq // 10
        if mem.duplex == '':
            raw.tx_freq = mem.freq // 10
        elif mem.duplex == 'split':
            raw.tx_freq = mem.offset // 10
        elif mem.duplex == '-':
            raw.tx_freq = (mem.freq - mem.offset) // 10
        elif mem.duplex == '+':
            raw.tx_freq = (mem.freq + mem.offset) // 10
        else:
            raise errors.RadioError('Unsupported duplex mode %r' % mem.duplex)

        raw.narrow = mem.mode == 'NFM'
        try:
            raw.high_power = self.POWER_LEVELS.index(mem.power)
        except ValueError:
            raw.high_power = 1
        raw.scan = mem.skip != 'S'

        raw.bcl_disabled = 1
        raw.digital = 0
        for item in mem.extra:
            if item.get_name() == 'bcl':
                raw.bcl_disabled = not item.value
            elif item.get_name() == 'encryption_key':
                idx = self._ENC_OPTS.index(item.value)
                if idx == 0:
                    raw.digital = 0
                else:
                    raw.digital = 1
                    raw.encryption_key = idx - 1
            elif item.get_name() == 'hopping' and not raw.digital:
                raw.hopping = item.value

        if raw.digital:
            self._set_no_tone(raw.tx_tone)
            self._set_no_tone(raw.rx_tone)
        elif mem.tmode == 'Tone':
            self._set_ctcss(mem.rtone, raw.tx_tone)
            self._set_no_tone(raw.rx_tone)
        elif mem.tmode == 'TSQL':
            self._set_ctcss(mem.ctone, raw.tx_tone)
            self._set_ctcss(mem.ctone, raw.rx_tone)
        elif mem.tmode == 'DTCS':
            # According to the documentation we should use ".rx_dtcs" because
            # "has_rx_dtcs" is set. However, the code does not match the docs,
            # and we have to use ".dtcs".
            self._set_dtcs(mem.dtcs, mem.dtcs_polarity[0], raw.tx_tone)
            self._set_dtcs(mem.dtcs, mem.dtcs_polarity[1], raw.rx_tone)
        elif mem.tmode == 'Cross':
            tx_tmode, rx_tmode = mem.cross_mode.split('->', 2)
            self._set_squelch(tx_tmode, mem.rtone, mem.dtcs,
                              mem.dtcs_polarity[0], raw.tx_tone)
            self._set_squelch(rx_tmode, mem.ctone, mem.rx_dtcs,
                              mem.dtcs_polarity[1], raw.rx_tone)


@directory.register
class BaofengW31D(BaofengDigital):
    """Baofeng W31D"""
    MODEL = 'W31D'
    VALID_BANDS = [(400000000, 480000000)]
    POWER_LEVELS = [
        chirp_common.PowerLevel("Low", watts=0.5),
        chirp_common.PowerLevel("High", watts=2.00),
    ]
    _serial_id = 0x04
    _proto = PROTO_D


@directory.register
class BaofengBFT20D(BaofengDigital):
    """Baofeng BF-T20D"""
    MODEL = 'BF-T20D'
    VALID_BANDS = [(400000000, 480000000)]
    POWER_LEVELS = [
        chirp_common.PowerLevel("Low", watts=0.5),
        chirp_common.PowerLevel("High", watts=2.00),
    ]
    _serial_id = 0x05
    _proto = PROTO_D



# Copyright 2019 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from tests import icom_clone_simulator
from chirp import memmap
from chirp.drivers import icw32
from builtins import bytes

import glob
import os
import unittest

from chirp import directory
directory.import_drivers()


class BaseIcomCloneTest():
    def setUp(self):
        self.radio = directory.get_radio(self.RADIO_IDENT)(None)
        self.simulator = icom_clone_simulator.FakeIcomRadio(self.radio)
        self.radio.set_pipe(self.simulator)

    def image_filename(self, filename):
        tests_base = os.path.dirname(os.path.abspath(__file__))
        return os.path.join(tests_base, 'images', filename)

    def load_from_test_image(self, filename):
        self.simulator.load_from_file(self.image_filename(filename))

    def image_version(self, filename):
        end = self.radio.get_memsize()

        with open(self.image_filename(filename), 'rb') as f:
            data = f.read()
        if data[-8:-6] == b'%02x' % self.radio.get_model()[0]:
            return 2
        elif data[end-16:end] == b'IcomCloneFormat3':
            return 3

    def test_sync_in(self):
        test_file = self.IMAGE_FILE
        self.load_from_test_image(test_file)
        self.radio.sync_in()

        img_ver = self.image_version(test_file)
        if img_ver == 2:
            endstring = b''.join(b'%02x' % ord(c)
                                 for c in self.radio._model[:2])
            self.assertEqual(endstring + b'0001', self.radio._mmap[-8:])
        elif img_ver == 3:
            self.assertEqual(b'IcomCloneFormat3', self.radio._mmap[-16:])
        elif img_ver is None:
            if isinstance(self.radio, icw32.ICW32ERadio):
                # The IC-W32E does some unclean things with the image size to
                # pass pre-metadata detection. It will fail here, so just
                # skip it.
                return
            self.assertEqual(self.radio.get_memsize(), len(self.radio._mmap))

    def test_sync_out(self):
        self.radio._mmap = memmap.MemoryMapBytes(
            bytes(b'\x00') * self.radio.get_memsize())
        start = self.radio._ranges[0][0]
        self.radio._mmap[start] = bytes(b'abcdefgh')
        self.radio.sync_out()
        self.assertEqual(b'abcdefgh', self.simulator._memory[start:start + 8])


class TestRawRadioData(unittest.TestCase):
    def test_get_payload(self):
        radio = directory.get_radio('Icom_IC-2730A')(None)
        payload = radio.get_payload(bytes(b'\x00\x10\xFE\x00'), True, True)
        self.assertEqual(b'\x00\x10\xFF\x0E\x00\xF2', payload)

        payload = radio.get_payload(bytes(b'\x00\x10\xFE\x00'), True, False)
        self.assertEqual(b'\x00\x10\xFF\x0E\x00', payload)

    def test_process_frame_payload(self):
        radio = directory.get_radio('Icom_IC-2730A')(None)
        data = radio.process_frame_payload(bytes(b'\x00\x10\xFF\x0E\x00'))
        self.assertEqual(b'\x00\x10\xFE\x00', data)


class TestAdapterMeta(type):
    def __new__(cls, name, parents, dct):
        return super(TestAdapterMeta, cls).__new__(cls, name, parents, dct)


test_file_glob = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                              '..', 'images',
                              'Icom_*.img')
for image_file in glob.glob(test_file_glob):
    base, _ext = os.path.splitext(os.path.basename(image_file))

    try:
        radio = directory.get_radio(base)
    except Exception:
        continue

    if radio._raw_frames:
        # The simulator does not behave like a raw radio
        continue

    class_name = 'Test_%s' % base
    sys.modules[__name__].__dict__[class_name] = \
        TestAdapterMeta(class_name,
                        (BaseIcomCloneTest, unittest.TestCase),
                        dict(RADIO_IDENT=base, IMAGE_FILE=image_file))



import os
import struct

import pytest

from tests.unit import base
from chirp.drivers import bf_t1
from chirp.drivers import uv5r
from chirp import settings


class FMRadioTest(base.BaseTest):
    IMAGE = None
    RADIO_CLASS = None
    MEMOBJ = 'element'
    SETTINGS_INDEX = (0, 'element')

    def setUp(self):
        if self.IMAGE is None:
            self.skipTest('Base test')

        img = os.path.join(os.path.dirname(__file__),
                           '..', 'images', self.IMAGE)
        self.radio = self.RADIO_CLASS(img)

        return super().setUp()

    def set_fmvfo_raw(self, value):
        """Set FM VFO as integer directly in memory"""
        val_bytes = struct.pack('>H', value)
        obj = self.radio._memobj
        for element in self.MEMOBJ.split('.'):
            obj = getattr(obj, element)
        obj.set_raw(val_bytes)

    def get_fmvfo_raw(self):
        """Get FM VFO as integer directly from memory"""
        obj = self.radio._memobj
        for element in self.MEMOBJ.split('.'):
            obj = getattr(obj, element)
        return int(obj)

    def check_fmvfo(self, expected):
        """Check FM VFO from settings

        expected should be something like 106.1
        """
        settings = self.radio.get_settings()
        index, name = self.SETTINGS_INDEX
        vfo = settings[index][name]
        self.assertEqual(expected, float(vfo.value))

    def set_fmvfo(self, value):
        """Set FM VFO via settings

        value is a float like 106.1
        """
        index, name = self.SETTINGS_INDEX
        settings = self.radio.get_settings()
        vfo = settings[index][name]
        vfo.value = value
        self.radio.set_settings(settings)

    def test_get_original(self):
        # 100.0 in original form would be:
        # 100.0 - 65 = 35.0 * 10 = 350 = 0x015E

        # Set it directly in our memory in old format
        self.set_fmvfo_raw(0x015E)

        # Make sure we get the expected value from settings
        self.check_fmvfo(100.0)

    def test_set_original(self):
        # Set it via settings
        self.set_fmvfo(106.1)

        # Make sure it's the expected value in memory, old format, since our
        # test image was in old format
        self.assertEqual(0x019B, self.get_fmvfo_raw())

        # Make sure we still get it back from settings as expected
        self.check_fmvfo(106.1)

    def test_get_2022(self):
        # 100.0 in original form would be:
        # 100.0 - 65 = 35.0 * 10 = 350 = 0x015E
        # New form is byte-swapped: 0x5E01

        # Set it directly in our memory in new format
        self.set_fmvfo_raw(0x5E01)

        # Make sure we get the expected value from settings
        self.check_fmvfo(100.0)

    def test_set_2022(self):
        # First set it in new format directly to 100.0
        self.set_fmvfo_raw(0x5E01)

        # Make sure we see it via settings as expected
        self.check_fmvfo(100.0)

        # Now set it via settings to a new value
        self.set_fmvfo(106.1)

        # Make sure it is in new form in the memory
        self.assertEqual(0x9B01, self.get_fmvfo_raw())

        # Make sure it is returned properly in settings
        self.check_fmvfo(106.1)

    def _test_edges(self, new=True):
        """Test the edges around the supported range.

        Make sure that we don't confuse old and new values on either side of
        the supported range as the wrong format.
        """
        if new:
            default = 0x5E01
        else:
            default = 0x015E
        for dial in range(50, 500, 10):
            if 65 <= dial <= 108:
                self.set_fmvfo(dial)
                self.check_fmvfo(dial)
            else:
                # Set this to a good default value so that we can set via
                # settings below.
                self.set_fmvfo_raw(default)

                # We should not be able to set this out-of-range value
                self.assertRaises(settings.InvalidValueError,
                                  self.set_fmvfo, dial)

                # Can't physically set this in memory
                if dial <= 65:
                    continue

                # Set it anyway in memory directly
                self.set_fmvfo_raw(int((dial - 65) * 10))

                # Make sure we do not interpret it incorrectly
                self.assertRaises(KeyError,
                                  self.check_fmvfo, dial)

    def test_edges_original(self):
        self._test_edges(False)

    def test_edges_2022(self):
        self._test_edges(True)


class TestBFT1(FMRadioTest):
    IMAGE = 'Baofeng_BF-T1.img'
    RADIO_CLASS = bf_t1.BFT1
    SETTINGS_INDEX = (1, 'fm_vfo')
    MEMOBJ = 'settings.fm_vfo'


@pytest.mark.skip('This radio needs fixing')
class TestUV5R(FMRadioTest):
    IMAGE = 'Baofeng_UV-5R.img'
    RADIO_CLASS = uv5r.BaofengUV5R
    SETTINGS_INDEX = (3, 'fm_presets')
    MEMOBJ = 'fm_presets'



#!/usr/bin/env python
#
#  cpep8.py - Check Python source files for PEP8 compliance.
#
# Copyright 2015  Zachary T Welch  <zach@mandolincreekfarm.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import warnings

import os
import sys
import argparse
import subprocess

# pep8 has a FutureWarning about nested sets. This isn't our problem, so
# squelch it here during import.
with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    import pep8


parser = argparse.ArgumentParser()
parser.add_argument("-a", "--all", action="store_true",
                    help="Check all files, ignoring blacklist")
parser.add_argument("-d", "--dir", action="append", default=['chirp', 'tests'],
                    help="Root directory of source tree")
parser.add_argument("-s", "--stats", action="store_true",
                    help="Only show statistics")
parser.add_argument("--strict", action="store_true",
                    help="Ignore listed exceptions")
parser.add_argument("-S", "--scan", action="store_true",
                    help="Scan for additional files")
parser.add_argument("-u", "--update", action="store_true",
                    help="Update manifest/blacklist files")
parser.add_argument("-v", "--verbose", action="store_true",
                    help="Display list of checked files")
parser.add_argument("files", metavar="file", nargs='*',
                    help="List of files to check (if none, check all)")
args = parser.parse_args()


def file_to_lines(name):
    fh = open(name, "r")
    lines = fh.read().split("\n")
    lines.pop()
    fh.close()
    return [x for x in lines if not x.startswith('#')]


scriptdir = os.path.dirname(sys.argv[0])
manifest_filename = os.path.join(scriptdir, "cpep8.manifest")
blacklist_filename = os.path.join(scriptdir, "cpep8.blacklist")
exceptions_filename = os.path.join(scriptdir, "cpep8.exceptions")

cpep8_manifest = set(file_to_lines(manifest_filename))
flake8_manifest = set()
for src_dir in [os.path.join('.', d) for d in args.dir]:
    for root, dirs, files in os.walk(src_dir):
        for f in files:
            filename = os.path.join(root, f)
            if filename.replace('\\', '/') in cpep8_manifest:
                continue
            if f.endswith('.py'):
                flake8_manifest.add(filename)
                continue
            with open(filename, "rb") as fh:
                shebang = fh.readline()
                if shebang.startswith(b"#!/usr/bin/env python"):
                    flake8_manifest.add(filename)


# unless we are being --strict, load per-file style exceptions
exceptions = {}
if not args.strict:
    exception_lines = file_to_lines(exceptions_filename)
    exception_lists = [x.split('\t')
                       for x in exception_lines if not x.startswith('#')]
    for filename, codes in exception_lists:
        exceptions[filename] = codes


def get_exceptions(f):
    try:
        ignore = exceptions[f]
    except KeyError:
        ignore = None
    return ignore


if args.files:
    old_files = cpep8_manifest
    cpep8_manifest = []
    flake8_manifest = []
    for fn in args.files:
        if not fn.startswith('./'):
            fn = './' + fn
        if fn in old_files:
            cpep8_manifest.append(fn)
        else:
            flake8_manifest.append(fn)

# read the blacklisted source files
blacklist = file_to_lines(blacklist_filename)

check_list = []
for f in cpep8_manifest:
    if args.all or f.replace('\\', '/') not in blacklist:
        check_list.append(f)
check_list = sorted(check_list)

total_errors = 0
for f in check_list:
    if args.verbose:
        print("Checking %s" % f)

    checker = pep8.Checker(f, quiet=args.stats, ignore=get_exceptions(f))
    results = checker.check_all()
    if args.stats:
        checker.report.print_statistics()
    total_errors += results

flake8_manifest = [f for f in flake8_manifest
                   if f.replace('\\', '/') not in blacklist]

for i in range(0, len(flake8_manifest), 10):
    files = flake8_manifest[i:i + 10]
    if args.verbose:
        print('Checking %s' % files)
    r = subprocess.call(['flake8', '--builtins=_,ngettext'] + files)
    if r != 0:
        total_errors += r

sys.exit(total_errors and 1 or 0)
